# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------
# The following codes are baed on those originally written by David B. Jess and Samuel D. T. Grant
# -----------------------------------------------------------------------------------------------------

import numpy as np # type: ignore
from matplotlib.colors import ListedColormap # type: ignore
from WaLSAtools import WaLSA_histo_opt # type: ignore
from scipy.interpolate import griddata # type: ignore
import matplotlib.pyplot as plt # type: ignore
from matplotlib.ticker import FixedLocator, FixedFormatter # type: ignore
import matplotlib.patches as patches # type: ignore
import matplotlib.patheffects as path_effects # type: ignore
from matplotlib.colors import Normalize # type: ignore

def WaLSA_plot_k_omega(
    kopower, kopower_xscale, kopower_yscale, 
    xtitle='Wavenumber (pixel⁻¹)', ytitle='Frequency (mHz)',
    xlog=False, ylog=False, xrange=None, yrange=None,
    xtick_interval=0.05, ytick_interval=200,
    xtick_minor_interval=0.01, ytick_minor_interval=50,
    colorbar_label='Log₁₀(Power)', cmap='viridis', cbartab=0.12,
    figsize=(9, 4), smooth=True, dypix = 1200, dxpix = 1600, ax=None,
    k1=None, k2=None, f1=None, f2=None, colorbar_location='right'
):
    """
    Plots the k-omega diagram with corrected orientation and axis alignment.
    Automatically clips the power array and scales based on x and y ranges.

    Example usage:
    WaLSA_plot_k_omega(
        kopower=power,
        kopower_xscale=wavenumber,
        kopower_yscale=frequencies*1000.,
        xlog=False, ylog=False, 
        xrange=(0, 0.3), figsize=(8, 4), cbartab=0.18,
        xtitle='Wavenumber (pixel⁻¹)', ytitle='Frequency (mHz)',
        colorbar_label='Log₁₀(Oscillation Power)',
        f1=0.470*1000, f2=0.530*1000,
        k1=0.047, k2=0.25
    )
    """

    if xrange is None or len(xrange) != 2:
        xrange = [np.min(kopower_xscale), np.max(kopower_xscale)]

    if yrange is None or len(yrange) != 2:
        yrange = [np.min(kopower_yscale), np.max(kopower_yscale)]

    # Handle xrange and yrange adjustments
    if xrange is not None and len(xrange) == 2 and xrange[0] == 0:
        xrange = [np.min(kopower_xscale), xrange[1]]

    if yrange is not None and len(yrange) == 2 and yrange[0] == 0:
        yrange = [np.min(kopower_yscale), yrange[1]]

    # Clip kopower, kopower_xscale, and kopower_yscale based on xrange and yrange
    if xrange:
        x_min, x_max = xrange
        x_indices = np.where((kopower_xscale >= x_min) & (kopower_xscale <= x_max))[0]

        # Check if the lower bound is not included
        if kopower_xscale[x_indices[0]] > x_min and x_indices[0] > 0:
            x_indices = np.insert(x_indices, 0, x_indices[0] - 1)

        # Check if the upper bound is not included
        if kopower_xscale[x_indices[-1]] < x_max and x_indices[-1] + 1 < len(kopower_xscale):
            x_indices = np.append(x_indices, x_indices[-1] + 1)

        kopower = kopower[:, x_indices]
        kopower_xscale = kopower_xscale[x_indices]

    if yrange:
        y_min, y_max = yrange
        y_indices = np.where((kopower_yscale >= y_min) & (kopower_yscale <= y_max))[0]

        # Check if the lower bound is not included
        if kopower_yscale[y_indices[0]] > y_min and y_indices[0] > 0:
            y_indices = np.insert(y_indices, 0, y_indices[0] - 1)

        # Check if the upper bound is not included
        if kopower_yscale[y_indices[-1]] < y_max and y_indices[-1] + 1 < len(kopower_yscale):
            y_indices = np.append(y_indices, y_indices[-1] + 1)

        kopower = kopower[y_indices, :]
        kopower_yscale = kopower_yscale[y_indices]

    # Pixel coordinates in data space
    if xlog:
        nxpix = np.logspace(np.log10(xrange[0]), np.log10(xrange[1]), dxpix)
    else:
        nxpix = np.linspace(xrange[0], xrange[1], dxpix)

    if ylog:
        nypix = np.logspace(np.log10(yrange[0]), np.log10(yrange[1]), dypix)
    else:
        nypix = np.linspace(yrange[0], yrange[1], dypix)

    # Interpolation over data
    interpolator = griddata(
        points=(np.repeat(kopower_yscale, len(kopower_xscale)), np.tile(kopower_xscale, len(kopower_yscale))),
        values=kopower.ravel(),
        xi=(nypix[:, None], nxpix[None, :]),
        method='linear'
    )
    newimage = np.nan_to_num(interpolator)

    # Clip the interpolated image to the exact xrange and yrange
    x_indices = (nxpix >= xrange[0]) & (nxpix <= xrange[1])
    y_indices = (nypix >= yrange[0]) & (nypix <= yrange[1])
    newimage = newimage[np.ix_(y_indices, x_indices)]
    nxpix = nxpix[x_indices]
    nypix = nypix[y_indices]

    # Extent for plotting
    extent = [nxpix[0], nxpix[-1], nypix[0], nypix[-1]]

    # Set up the plot
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    # Load custom colormap
    rgb_values = np.loadtxt('Color_Tables/idl_colormap_13.txt') / 255.0
    idl_colormap_13 = ListedColormap(rgb_values)

    # Compute minimum and maximum values of the data
    vmin = np.nanmin(kopower) 
    vmax = np.nanmax(kopower) 

    img = ax.imshow(
        WaLSA_histo_opt(newimage), extent=extent, origin='lower', aspect='auto', cmap=idl_colormap_13, norm=Normalize(vmin=vmin, vmax=vmax)
    )

    # Configure axis labels and scales
    ax.set_xlabel(xtitle)
    ax.set_ylabel(ytitle)
    if xlog:
        ax.set_xscale('log')
    if ylog:
        ax.set_yscale('log')

    # Configure ticks
    ax.xaxis.set_major_locator(plt.MultipleLocator(xtick_interval))
    ax.xaxis.set_minor_locator(plt.MultipleLocator(xtick_minor_interval))
    ax.yaxis.set_major_locator(plt.MultipleLocator(ytick_interval))
    ax.yaxis.set_minor_locator(plt.MultipleLocator(ytick_minor_interval))

    # # Add a secondary x and y axes for perido and spatial size
    ax.tick_params(axis='both', which='both', direction='out', top=True, right=True)
    ax.tick_params(axis='both', which='major', length=7, width=1.1)  # Major ticks
    ax.tick_params(axis='both', which='minor', length=4, width=1.1)  # Minor ticks

    major_xticks = ax.get_xticks()
    major_yticks = ax.get_yticks()
    
    # Define a function to calculate the period (1000 / frequency)
    def frequency_to_period(frequency):
        return 1000. / frequency if frequency > 0 else np.inf
    # Generate labels for the secondary y-axis based on the primary y-axis ticks
    period_labels = [f"{frequency_to_period(tick):.1f}" if tick > 0 else "" for tick in major_yticks]

    # Set custom labels for the secondary y-axis
    ax_right = ax.secondary_yaxis('right')
    ax_right.set_yticks(major_yticks)
    ax_right.set_yticklabels(period_labels)
    ax_right.set_ylabel('Period (s)', labelpad=12)

    # Define a function to calculate spatial size (2π / wavenumber)
    def wavenumber_to_spatial_size(wavenumber):
        return 2 * np.pi / wavenumber if wavenumber > 0 else np.inf

    # Generate labels for the secondary x-axis based on the primary x-axis ticks
    spatial_size_labels = [f"{wavenumber_to_spatial_size(tick):.1f}" if tick > 0 else "" for tick in major_xticks]

    # Set custom labels for the secondary x-axis
    ax_top = ax.secondary_xaxis('top')
    ax_top.set_xticks(major_xticks)
    ax_top.set_xticklabels(spatial_size_labels)
    ax_top.set_xlabel('Spatial Size (pixel)', labelpad=12)

    if k1 is not None and k2 is not None and f1 is not None and f2 is not None:
        width = k2 - k1
        height = f2 - f1
        rectangle = patches.Rectangle(
            (k1, f1), width, height, zorder=10,
            linewidth=1.5, edgecolor='white', facecolor='none', linestyle='--'
        )
        rectangle.set_path_effects([
            path_effects.Stroke(linewidth=2.5, foreground='black'),
            path_effects.Normal()
        ])
        # Mark the filtered area
        ax.add_patch(rectangle)

    # Add colorbar
    tick_values = [
        vmin,
        vmin + (vmax - vmin) * 0.33,
        vmin + (vmax - vmin) * 0.67,
        vmax
    ]
    # Format the tick labels
    tick_labels = [f"{v:.1f}" for v in tick_values] 

    if colorbar_location == 'top':
        cbar = plt.colorbar(img, ax=ax, orientation='horizontal', pad=cbartab, location='top', aspect=30)
        cbar.set_label(colorbar_label)
        # Override the ticks and tick labels
        cbar.set_ticks(tick_values)       # Set custom tick locations
        cbar.set_ticklabels(tick_labels)  # Set custom tick labels
        cbar.ax.xaxis.set_major_locator(FixedLocator(tick_values))  # Fix tick locations
        cbar.ax.xaxis.set_major_formatter(FixedFormatter(tick_labels))  # Fix tick labels
        # Suppress auto ticks completely
        cbar.ax.xaxis.set_minor_locator(FixedLocator([]))  # Ensure no minor ticks appear
    else:
        cbar = plt.colorbar(img, ax=ax, orientation='vertical', pad=cbartab, aspect=22)
        cbar.set_label(colorbar_label)
        # Override the ticks and tick labels
        cbar.set_ticks(tick_values)       # Set custom tick locations
        cbar.set_ticklabels(tick_labels)  # Set custom tick labels
        cbar.ax.yaxis.set_major_locator(FixedLocator(tick_values))  # Fix tick locations
        cbar.ax.yaxis.set_major_formatter(FixedFormatter(tick_labels))  # Fix tick labels
        # Suppress auto ticks completely
        cbar.ax.yaxis.set_minor_locator(FixedLocator([]))  # Ensure no minor ticks appear

    return ax


# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

import subprocess
import shutil
import os
import stat
import numpy as np # type: ignore
from matplotlib.backends.backend_pdf import PdfPages # type: ignore

def WaLSA_save_pdf(fig, pdf_path, color_mode='RGB', dpi=300, bbox_inches=None, pad_inches=0):
    """
    Save a PDF from a Matplotlib figure with an option to convert to CMYK if Ghostscript is available.
    
    Parameters:
    - fig: Matplotlib figure object to save.
    - pdf_path: Path to save the PDF file.
    - color_mode: 'RGB' (default) to save in RGB, or 'CMYK' to convert to CMYK.
    - dpi: Resolution in dots per inch. Default is 300.
    - bbox_inches: Set to 'tight' to remove extra white space. Default is 'tight'.
    - pad_inches: Padding around the figure. Default is 0.
    """

    # Save the initial PDF in RGB format using fig.savefig to apply bbox_inches, pad_inches, and dpi
    if bbox_inches is None:
        # Save the initial PDF in RGB format
        with PdfPages(pdf_path) as pdf:
            pdf.savefig(fig, transparent=True)

        if color_mode.upper() == 'CMYK':
            # Check if Ghostscript is installed
            if shutil.which("gs") is not None:
                # Set permissions on the initial RGB PDF to ensure access for Ghostscript
                os.chmod(pdf_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)

                # Define a temporary path for the CMYK file
                temp_cmyk_path = pdf_path.replace('.pdf', '_temp_cmyk.pdf')

                # Run Ghostscript to create the CMYK PDF as a temporary file
                subprocess.run([
                    "gs", "-o", temp_cmyk_path,
                    "-sDEVICE=pdfwrite",
                    "-dProcessColorModel=/DeviceCMYK",
                    "-dColorConversionStrategy=/CMYK",
                    "-dNOPAUSE", "-dBATCH", "-dSAFER",
                    "-dPDFSETTINGS=/prepress", pdf_path
                ], check=True)

                # Replace the original RGB PDF with the CMYK version
                os.remove(pdf_path)  # Delete the RGB version
                os.rename(temp_cmyk_path, pdf_path)  # Rename CMYK as the original file
                print(f"PDF saved in CMYK format as '{pdf_path}'")
            
            else:
                print("Warning: Ghostscript is not installed, so the PDF remains in RGB format.")
                print("To enable CMYK saving, please install Ghostscript:")
                print("- On macOS, use: brew install ghostscript")
                print("- On Ubuntu, use: sudo apt install ghostscript")
                print("- On Windows, download from: https://www.ghostscript.com/download.html")
        else:
            print(f"PDF saved in RGB format as '{pdf_path}'")

    else:
        # Temporary path for the RGB version of the PDF
        temp_pdf_path = pdf_path.replace('.pdf', '_temp.pdf')

        fig.savefig(
            temp_pdf_path, 
            format='pdf', 
            dpi=dpi, 
            bbox_inches=bbox_inches, 
            pad_inches=pad_inches, 
            transparent=True
        )

        if color_mode.upper() == 'CMYK':
            # Check if Ghostscript is installed
            if shutil.which("gs") is not None:
                # Set permissions on the initial RGB PDF to ensure access for Ghostscript
                os.chmod(temp_pdf_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)

                # Run Ghostscript to create the CMYK PDF as a temporary file
                subprocess.run([
                    "gs", "-o", pdf_path,
                    "-sDEVICE=pdfwrite",
                    "-dProcessColorModel=/DeviceCMYK",
                    "-dColorConversionStrategy=/CMYK",
                    "-dNOPAUSE", "-dBATCH", "-dSAFER",
                    "-dPDFSETTINGS=/prepress", temp_pdf_path
                ], check=True)

                # Remove the temporary PDF file
                os.remove(temp_pdf_path)
                print(f"PDF saved in CMYK format as '{pdf_path}'")
            
            else:
                print("Warning: Ghostscript is not installed, so the PDF remains in RGB format.")
                print("To enable CMYK saving, please install Ghostscript:")
                print("- On macOS, use: brew install ghostscript")
                print("- On Ubuntu, use: sudo apt install ghostscript")
                print("- On Windows, download from: https://www.ghostscript.com/download.html")
        else:
            # Rename the temporary file as the final file if no CMYK conversion is needed
            os.rename(temp_pdf_path, pdf_path)
            print(f"PDF saved in RGB format as '{pdf_path}'")


def WaLSA_histo_opt(image, cutoff=1e-3, top_only=False, bot_only=False):
    """
    Clip image values based on the cutoff percentile to enhance contrast.
    Inspired by IDL's "iris_histo_opt" function (Copyright: P.Suetterlin, V.Hansteen, and M. Carlsson)
    
    Parameters:
    - image: 2D array (image data).
    - cutoff: Fraction of values to clip at the top and bottom (default is 0.001).
    - top_only: If True, clip only the highest values.
    - bot_only: If True, clip only the lowest values.
    
    Returns:
    - Clipped image for better contrast.
    """
    # Ignore NaNs in the image
    finite_values = image[np.isfinite(image)]
    
    # Calculate lower and upper bounds based on cutoff percentiles
    lower_bound = np.percentile(finite_values, cutoff * 100)
    upper_bound = np.percentile(finite_values, (1 - cutoff) * 100)
    
    # Clip image according to bounds and options
    if top_only:
        return np.clip(image, None, upper_bound)
    elif bot_only:
        return np.clip(image, lower_bound, None)
    else:
        return np.clip(image, lower_bound, upper_bound)
    


# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

import numpy as np # type: ignore

def WaLSA_confidence(ps_perm, siglevel=0.05, nf=None):
    """
    Find the confidence levels (significance levels) for the given power spectrum permutations.

    Parameters:
    -----------
    ps_perm : np.ndarray
        2D array where each row represents a power spectrum, and each column represents a permutation.
    siglevel : float
        Significance level (default 0.05 for 95% confidence).
    nf : int
        Number of frequencies (length of the power spectra). If None, inferred from the shape of ps_perm.

    Returns:
    --------
    signif : np.ndarray
        Significance levels for each frequency.
    """
    if nf is None:
        nf = ps_perm.shape[0]  # Number of frequencies inferred from the shape of ps_perm

    signif = np.zeros((nf))

    # Loop through each frequency
    for iv in range(nf):
        # Extract the permutation values for this frequency
        tmp = np.sort(ps_perm[iv, :])  # Sort the power spectrum permutation values

        # Calculate the number of permutations
        ntmp = len(tmp)

        # Find the significance threshold
        nsig = int(round(siglevel * ntmp))  # Number of permutations to cut off for the significance level

        # Set the confidence level for this frequency
        signif[iv] = tmp[-nsig]  # Select the (ntmp - nsig)-th element (equivalent to IDL's ROUND and indexing)

    return signif



#!/usr/bin/python
# coding: UTF-8
#
# Author:   Dawid Laszuk
# Contact:  https://github.com/laszukdawid/PyEMD/issues
#
# Feel free to contact for any information.
"""
.. currentmodule:: EEMD
"""

import logging
from collections import defaultdict
from multiprocessing import Pool
from typing import Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from tqdm import tqdm

from .utils import get_timeline


class EEMD:
    """
    **Ensemble Empirical Mode Decomposition**

    Ensemble empirical mode decomposition (EEMD) [Wu2009]_
    is noise-assisted technique, which is meant to be more robust
    than simple Empirical Mode Decomposition (EMD). The robustness is
    checked by performing many decompositions on signals slightly
    perturbed from their initial position. In the grand average over
    all IMF results the noise will cancel each other out and the result
    is pure decomposition.

    Parameters
    ----------
    trials : int (default: 100)
        Number of trials or EMD performance with added noise.
    noise_width : float (default: 0.05)
        Standard deviation of Gaussian noise ().
        It's relative to absolute amplitude of the signal, i.e.
        :math:``, where
        :math:`sigma` is noise_width.
    ext_EMD : EMD (default: None)
        One can pass EMD object defined outside, which will be
        used to compute IMF decompositions in each trial. If none
        is passed then EMD with default options is used.
    parallel : bool (default: False)
        Flag whether to use multiprocessing in EEMD execution.
        Since each EMD(s+noise) is independent this should improve execution
        speed considerably.
        *Note* that it's disabled by default because it's the most common
        problem when EEMD takes too long time to finish.
        If you set the flag to True, make also sure to set `processes` to
        some reasonable value.
    processes : int or None (optional)
        Number of processes harness when executing in parallel mode.
        The value should be between 1 and max that depends on your hardware.
    separate_trends : bool (default: False)
        Flag whether to isolate trends from each EMD decomposition into a separate component.
        If `true`, the resulting EEMD will contain ensemble only from IMFs and
        the mean residue will be stacked as the last element.

    References
    ----------
    .. [Wu2009] Z. Wu and N. E. Huang, "Ensemble empirical mode decomposition:
        A noise-assisted data analysis method", Advances in Adaptive
        Data Analysis, Vol. 1, No. 1 (2009) 1-41.
    """

    logger = logging.getLogger(__name__)

    noise_kinds_all = ["normal", "uniform"]

    def __init__(self, trials: int = 100, noise_width: float = 0.05, ext_EMD=None, parallel: bool = False, **kwargs):
        # Ensemble constants
        self.trials = trials
        self.noise_width = noise_width
        self.separate_trends = bool(kwargs.get("separate_trends", False))

        self.random = np.random.RandomState()
        self.noise_kind = kwargs.get("noise_kind", "normal")
        self.parallel = parallel
        self.processes = kwargs.get("processes")  # Optional[int]
        if self.processes is not None and not self.parallel:
            self.logger.warning("Passed value for process has no effect when `parallel` is False.")

        if ext_EMD is None:
            from .EMD import EMD

            self.EMD = EMD(**kwargs)
        else:
            self.EMD = ext_EMD

        self.E_IMF = None  # Optional[np.ndarray]
        self.residue = None  # Optional[np.ndarray]
        self._all_imfs = {}

    def __call__(
        self, S: np.ndarray, T: Optional[np.ndarray] = None, max_imf: int = -1, progress: bool = False
    ) -> np.ndarray:
        return self.eemd(S, T=T, max_imf=max_imf, progress=progress)

    def __getstate__(self) -> Dict:
        self_dict = self.__dict__.copy()
        return self_dict

    def generate_noise(self, scale: float, size: Union[int, Sequence[int]]) -> np.ndarray:
        """
        Generate noise with specified parameters.
        Currently supported distributions are:

        * *normal* with std equal scale.
        * *uniform* with range [-scale/2, scale/2].

        Parameters
        ----------
        scale : float
            Width for the distribution.
        size : int
            Number of generated samples.

        Returns
        -------
        noise : numpy array
            Noise sampled from selected distribution.
        """
        if self.noise_kind == "normal":
            noise = self.random.normal(loc=0, scale=scale, size=size)
        elif self.noise_kind == "uniform":
            noise = self.random.uniform(low=-scale / 2, high=scale / 2, size=size)
        else:
            raise ValueError(
                "Unsupported noise kind. Please assigned `noise_kind` to be one of these: {0}".format(
                    str(self.noise_kinds_all)
                )
            )
        return noise

    def noise_seed(self, seed: int) -> None:
        """Set seed for noise generation."""
        self.random.seed(seed)

    def eemd(
        self, S: np.ndarray, T: Optional[np.ndarray] = None, max_imf: int = -1, progress: bool = False
    ) -> np.ndarray:
        """
        Performs EEMD on provided signal.

        For a large number of iterations defined by `trials` attr
        the method performs :py:meth:`emd` on a signal with added white noise.

        Parameters
        ----------
        S : numpy array,
            Input signal on which EEMD is performed.
        T : numpy array or None, (default: None)
            If none passed samples are numerated.
        max_imf : int, (default: -1)
            Defines up to how many IMFs each decomposition should
            be performed. By default (negative value) it decomposes
            all IMFs.

        Returns
        -------
        eIMF : numpy array
            Set of ensemble IMFs produced from input signal. In general,
            these do not have to be, and most likely will not be, same as IMFs
            produced using EMD.
        """
        if T is None:
            T = get_timeline(len(S), S.dtype)

        scale = self.noise_width * np.abs(np.max(S) - np.min(S))
        self._S = S
        self._T = T
        self._N = len(S)
        self._scale = scale
        self.max_imf = max_imf

        # For trial number of iterations perform EMD on a signal
        # with added white noise
        if self.parallel:
            pool = Pool(processes=self.processes)
            map_pool = pool.map
        else:
            map_pool = map
        all_IMFs = map_pool(self._trial_update, range(self.trials))

        if self.parallel:
            pool.close()

        self._all_imfs = defaultdict(list)
        it = iter if not progress else lambda x: tqdm(x, desc="EEMD", total=self.trials)
        for imfs, trend in it(all_IMFs):
            # A bit of explanation here.
            # If the `trend` is not None, that means it was intentionally separated in the decomp process.
            # This might due to `separate_trends` flag which means that trends are summed up separately
            # and treated as the last component. Since `proto_eimfs` is a dict, that `-1` is treated literally
            # and **not** as the *last position*. We can then use that `-1` to always add it as the last pos
            # in the actual eIMF, which indicates the trend.
            if trend is not None:
                self._all_imfs[-1].append(trend)

            for imf_num, imf in enumerate(imfs):
                self._all_imfs[imf_num].append(imf)

        # Convert defaultdict back to dict and explicitly rename `-1` position to be {the last value} for consistency.
        self._all_imfs = dict(self._all_imfs)
        if -1 in self._all_imfs:
            self._all_imfs[len(self._all_imfs)] = self._all_imfs.pop(-1)

        for imf_num in self._all_imfs.keys():
            self._all_imfs[imf_num] = np.array(self._all_imfs[imf_num])

        self.E_IMF = self.ensemble_mean()
        self.residue = S - np.sum(self.E_IMF, axis=0)

        return self.E_IMF

    def _trial_update(self, trial) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """A single trial evaluation, i.e. EMD(signal + noise).

        *Note*: Although `trial` argument isn't used it's needed for the (multiprocessing) map method.
        """
        noise = self.generate_noise(self._scale, self._N)
        imfs = self.emd(self._S + noise, self._T, self.max_imf)
        trend = None
        if self.separate_trends:
            imfs, trend = self.EMD.get_imfs_and_trend()

        return (imfs, trend)

    def emd(self, S: np.ndarray, T: np.ndarray, max_imf: int = -1) -> np.ndarray:
        """Vanilla EMD method.

        Provides emd evaluation from provided EMD class.
        For reference please see :class:`PyEMD.EMD`.
        """
        return self.EMD.emd(S, T, max_imf=max_imf)

    def get_imfs_and_residue(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Provides access to separated imfs and residue from recently analysed signal.

        Returns
        -------
        (imfs, residue) : (np.ndarray, np.ndarray)
            Tuple that contains all imfs and a residue (if any).

        """
        if self.E_IMF is None or self.residue is None:
            raise ValueError("No IMF found. Please, run EMD method or its variant first.")
        return self.E_IMF, self.residue

    @property
    def all_imfs(self):
        """A dictionary with all computed imfs per given order."""
        return self._all_imfs

    def ensemble_count(self) -> List[int]:
        """Count of imfs observed for given order, e.g. 1st proto-imf, in the whole ensemble."""
        return [len(imfs) for imfs in self._all_imfs.values()]

    def ensemble_mean(self) -> np.ndarray:
        """Pointwise mean over computed ensemble. Same as the output of `eemd()` method."""
        return np.array([imfs.mean(axis=0) for imfs in self._all_imfs.values()])

    def ensemble_std(self) -> np.ndarray:
        """Pointwise standard deviation over computed ensemble."""
        return np.array([imfs.std(axis=0) for imfs in self._all_imfs.values()])


###################################################
# Beginning of program
if __name__ == "__main__":
    import pylab as plt

    E_imfNo = np.zeros(50, dtype=np.int)

    # Logging options
    logging.basicConfig(level=logging.INFO)

    # EEMD options
    max_imf = -1

    # Signal options
    N = 500
    tMin, tMax = 0, 2 * np.pi
    T = np.linspace(tMin, tMax, N)

    S = 3 * np.sin(4 * T) + 4 * np.cos(9 * T) + np.sin(8.11 * T + 1.2)

    # Prepare and run EEMD
    eemd = EEMD(trials=50)
    eemd.noise_seed(12345)

    E_IMFs = eemd.eemd(S, T, max_imf)
    imfNo = E_IMFs.shape[0]

    # Plot results in a grid
    c = np.floor(np.sqrt(imfNo + 1))
    r = np.ceil((imfNo + 1) / c)

    plt.ioff()
    plt.subplot(r, c, 1)
    plt.plot(T, S, "r")
    plt.xlim((tMin, tMax))
    plt.title("Original signal")

    for num in range(imfNo):
        plt.subplot(r, c, num + 2)
        plt.plot(T, E_IMFs[num], "g")
        plt.xlim((tMin, tMax))
        plt.title("Imf " + str(num + 1))

    plt.show()



# -----------------------------------------------------------------------------------------------------
# WaLSAtools - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

import numpy as np # type: ignore
from scipy.signal import coherence, csd # type: ignore
from .WaLSA_speclizer import WaLSA_speclizer # type: ignore
from .WaLSA_wavelet import xwt, wct # type: ignore
from .WaLSA_detrend_apod import WaLSA_detrend_apod # type: ignore
from .WaLSA_wavelet_confidence import WaLSA_wavelet_confidence # type: ignore

# -------------------------------------- Main Function ----------------------------------------
def WaLSA_cross_spectra(signal=None, time=None, method=None, **kwargs):
    """
    Compute the cross-spectra between two time series.
    
    Parameters:
        data1 (array): The first time series (1D).
        data2 (array): The first time series (1D).
        time (array): The time array of the signals.
        method (str): The method to use for the analysis. Options are 'welch' and 'wavelet'.
    """
    if method == 'welch':
        return getcross_spectrum_Welch(signal, time=time, **kwargs)
    elif method == 'wavelet':
        return getcross_spectrum_Wavelet(signal, time=time, **kwargs)
    else:
        raise ValueError(f"Unknown method: {method}")


# -------------------------------------- Welch ----------------------------------------
def getcross_spectrum_Welch(signal, time, **kwargs):
    """
    Calculate cross-spectral relationships of two time series whose amplitudes and powers
    are computed using the WaLSA_speclizer routine.
    The cross-spectrum is complex valued, thus its magnitude is returned as the
    co-spectrum. The phase lags between the two time series are are estimated 
    from the imaginary and real arguments of the complex cross spectrum.
    The coherence is calculated from the normalized square of the amplitude 
    of the complex cross-spectrum
    
    Parameters:
        data1 (array): The first 1D time series signal.
        data2 (array): The second 1D time series signal.
        time (array): The time array corresponding to the signals.
        nperseg (int, optional): Length of each segment for analysis. Default: 256.
        noverlap (int, optional): Number of points to overlap between segments. Default: 128.
        window (str, optional): Type of window function used in the Welch method. Default: 'hann'.
        siglevel (float, optional): Significance level for confidence intervals. Default: 0.95.
        nperm (int, optional): Number of permutations for significance testing. Default: 1000.
        silent (bool, optional): If True, suppress print statements. Default: False.
        **kwargs: Additional parameters for the analysis method.
    
    Returns:
        cospectrum, frequencies, phase_angle, coherence, signif_cross, signif_coh, d1_power, d2_power
    """
    # Define default values for the optional parameters
    defaults = {
        'data1': None,        # First data array
        'data1': None,        # Second data array
        'nperseg': 256,      # Number of points per segments to average
        'apod': 0.1,       # Apodization function
        'nodetrendapod': None,  # No detrending ot apodization applied
        'pxdetrend': 2,  # Detrend parameter
        'meandetrend': None,  # Detrend parameter
        'polyfit': None,    # Detrend parameter
        'meantemporal': None,  # Detrend parameter
        'recon': None      # Detrend parameter
    }

    # Update defaults with any user-provided keyword arguments
    params = {**defaults, **kwargs}

    data1 = params['data1']
    data2 = params['data2']

    dummy = signal+2

    tdiff = np.diff(time)
    cadence = np.median(tdiff)
    
    # Power spectrum for data1
    power_data1, frequencies, _ = WaLSA_speclizer(signal=data1, time=time, method='welch', 
                                               amplitude=True, nosignificance=True, silent=kwargs.pop('silent', True), **kwargs)

    # Power spectrum for data2
    power_data2, _, _ = WaLSA_speclizer(signal=data2, time=time, method='welch', 
                                               amplitude=True, nosignificance=True, silent=kwargs.pop('silent', False), **kwargs)

    # Calculate cross-spectrum
    _, crosspower = csd(data1, data2, fs=1.0/cadence, window='hann', nperseg=500)

    cospectrum = np.abs(crosspower)
    
    # Calculate phase lag
    phase_angle = np.angle(crosspower, deg=True)

    # Calculate coherence
    freq_coh, coh = coherence(data1, data2, 1.0/cadence, nperseg=500)
    
    return frequencies, cospectrum, phase_angle, power_data1, power_data2, freq_coh, coh


# -------------------------------------- Wavelet ----------------------------------------
def getcross_spectrum_Wavelet(signal, time, **kwargs):
    """
    Calculate the cross-power spectrum, coherence spectrum, and phase angles 
    between two signals using Wavelet Transform.

    Parameters:
        data1 (array): The first 1D time series signal.
        data2 (array): The second 1D time series signal.
        time (array): The time array corresponding to the signals.
        siglevel (float): Significance level for the confidence intervals. Default: 0.95.
        nperm (int): Number of permutations for significance testing. Default: 1000.
        mother (str): The mother wavelet function to use. Default: 'morlet'.
        GWS (bool): If True, calculate the Global Wavelet Spectrum. Default: False.
        RGWS (bool): If True, calculate the Refined Global Wavelet Spectrum (time-integrated power, excluding COI and insignificant areas). Default: False.
        dj (float): Scale spacing. Smaller values result in better scale resolution but slower calculations. Default: 0.025.
        s0 (float): Initial (smallest) scale of the wavelet. Default: 2 * dt.
        J (int): Number of scales minus one. Scales range from s0 up to s0 * 2**(J * dj), giving a total of (J + 1) scales. Default: (log2(N * dt / s0)) / dj.
        lag1 (float): Lag-1 autocorrelation. Default: 0.0.
        apod (float): Extent of apodization edges (of a Tukey window). Default: 0.1.
        pxdetrend (int): Subtract linear trend with time per pixel. Options: 1 (simple) or 2 (advanced). Default: 2.
        polyfit (int): Degree of polynomial fit for detrending the data. If set, a polynomial fit (instead of linear) is applied. Default: None.
        meantemporal (bool): If True, apply simple temporal detrending by subtracting the mean signal from the data, skipping fitting procedures. Default: False.
        meandetrend (bool): If True, subtract the linear trend with time for the image means (spatial detrending). Default: False.
        recon (bool): If True, perform Fourier reconstruction of the input time series. This does not preserve amplitudes but is useful for examining frequencies far from the low-frequency range.  Default: False.
        resample_original (bool): If True, and if recon set True, approximate values close to the original are returned for comparison. Default: False.
        nodetrendapod (bool): If True, neither detrending nor apodization is performed. Default: False.
        silent (bool): If True, suppress print statements. Default: False.

    Returns:
    cross_power : np.ndarray
        Cross power spectrum between `data1` and `data2`.
        
    cross_periods : np.ndarray
        Periods corresponding to the cross-power spectrum.
        
    cross_sig : np.ndarray
        Significance of the cross-power spectrum.
        
    cross_coi : np.ndarray
        Cone of influence for the cross-power spectrum.
        
    coherence : np.ndarray
        Coherence spectrum between `data1` and `data2`.
        
    coh_periods : np.ndarray
        Periods corresponding to the coherence spectrum.
        
    coh_sig : np.ndarray
        Significance of the coherence spectrum.
        
    corr_coi : np.ndarray
        Cone of influence for the coherence spectrum.
        
    phase_angle : np.ndarray
        2D array containing x- and y-components of the phase direction 
        arrows for each frequency and time point.
        
    Notes
    -----
    - Cross-power, coherence, and phase angles are calculated using 
      **cross-wavelet transform (XWT)** and **wavelet coherence transform (WCT)**.
    - Arrows for phase direction are computed such that:
        - Arrows pointing downwards indicate anti-phase.
        - Arrows pointing upwards indicate in-phase.
        - Arrows pointing right indicate `data1` leading `data2`.
        - Arrows pointing left indicate `data2` leading `data1`.

    Examples
    --------
    >>> cross_power, cross_periods, cross_sig, cross_coi, coherence, coh_periods, coh_sig, corr_coi, phase_angle, dt = \
    >>> getcross_spectrum_Wavelet(signal, cadence, data1=signal1, data2=signal2)
    """

    # Define default values for optional parameters
    defaults = {
        'data1': None,               # First data array
        'data2': None,                # Second data array
        'mother': 'morlet',          # Type of mother wavelet
        'siglevel': 0.95,  # Significance level for cross-spectral analysis
        'dj': 0.025,                 # Spacing between scales
        's0': -1,  # Initial scale
        'J': -1,  # Number of scales
        'cache': False,               # Cache results to avoid recomputation
        'apod': 0.1,
        'silent': False,
        'pxdetrend': 2,
        'meandetrend': False,
        'polyfit': None,
        'meantemporal': False,
        'recon': False,
        'resample_original': False,
        'nperm': 1000               # Number of permutations for significance testing
    }

    # Update defaults with any user-provided keyword arguments
    params = {**defaults, **kwargs}

    data1 = params['data1']
    data2 = params['data2']
    dummy = signal+2

    tdiff = np.diff(time)
    cadence = np.median(tdiff)

    data1orig = data1.copy()
    data2orig = data2.copy()

    data1 = WaLSA_detrend_apod(
        data1, 
        apod=params['apod'], 
        meandetrend=params['meandetrend'], 
        pxdetrend=params['pxdetrend'], 
        polyfit=params['polyfit'], 
        meantemporal=params['meantemporal'], 
        recon=params['recon'], 
        cadence=cadence, 
        resample_original=params['resample_original'], 
        silent=True
    )

    data2 = WaLSA_detrend_apod(
        data2, 
        apod=params['apod'], 
        meandetrend=params['meandetrend'], 
        pxdetrend=params['pxdetrend'], 
        polyfit=params['polyfit'], 
        meantemporal=params['meantemporal'], 
        recon=params['recon'], 
        cadence=cadence, 
        resample_original=params['resample_original'], 
        silent=True
    )

    # Calculate signal properties
    nt = len(params['data1'])  # Number of time points

    # Determine the initial scale s0 if not provided
    if params['s0'] == -1:
        params['s0'] = 2 * cadence

    # Determine the number of scales J if not provided
    if params['J'] == -1:
        params['J'] = int((np.log(float(nt) * cadence / params['s0']) / np.log(2)) / params['dj'])

    # ----------- CROSS-WAVELET TRANSFORM (XWT) -----------
    W12, cross_coi, freq, signif = xwt(
        data1, data2, cadence, dj=params['dj'], s0=params['s0'], J=params['J'], 
        no_default_signif=True, wavelet=params['mother'], normalize=False
    )
    print('Wavelet cross-power spectrum calculated.')
    # Calculate the cross-power spectrum
    cross_power = np.abs(W12) ** 2
    cross_periods = 1 / freq  # Periods corresponding to the frequency axis
    
    #----------------------------------------------------------------------
    # Calculate significance levels using Monte Carlo randomization method
    #----------------------------------------------------------------------
    nxx, nyy = cross_power.shape
    cross_perm = np.zeros((nxx, nyy, params['nperm']))
    print('\nCalculating wavelet cross-power significance:')
    total_iterations = params['nperm']  # Total number of (x, y) combinations
    iteration_count = 0  # To keep track of progress
    for ip in range(params['nperm']):
        iteration_count += 1
        progress = (iteration_count / total_iterations) * 100
        print(f"\rProgress: {progress:.2f}%", end="")

        y_perm1 = np.random.permutation(data1orig)  # Permuting the original signal
        apocube1 = WaLSA_detrend_apod(
            y_perm1, 
            apod=params['apod'], 
            meandetrend=params['meandetrend'], 
            pxdetrend=params['pxdetrend'], 
            polyfit=params['polyfit'], 
            meantemporal=params['meantemporal'], 
            recon=params['recon'], 
            cadence=cadence, 
            resample_original=params['resample_original'], 
            silent=True
        )
        y_perm2 = np.random.permutation(data2orig)  # Permuting the original signal
        apocube2 = WaLSA_detrend_apod(
            y_perm2, 
            apod=params['apod'], 
            meandetrend=params['meandetrend'], 
            pxdetrend=params['pxdetrend'], 
            polyfit=params['polyfit'], 
            meantemporal=params['meantemporal'], 
            recon=params['recon'], 
            cadence=cadence, 
            resample_original=params['resample_original'], 
            silent=True
        )
        W12s, _, _, _ = xwt(
            apocube1, apocube2, cadence, dj=params['dj'], s0=params['s0'], J=params['J'], 
            no_default_signif=True, wavelet=params['mother'], normalize=False
        )
        cross_power_sig = np.abs(W12s) ** 2
        cross_perm[:, :, ip] = cross_power_sig

    signifn = WaLSA_wavelet_confidence(cross_perm, siglevel=params['siglevel'])

    cross_sig = cross_power / signifn

    # ----------- WAVELET COHERENCE TRANSFORM (WCT) -----------
    WCT, aWCT, coh_coi, freq, sig = wct(
        data1, data2, cadence, dj=params['dj'], s0=params['s0'], J=params['J'], 
        sig=False, wavelet=params['mother'], normalize=False, cache=params['cache']
    )
    print('Wavelet coherence calculated.')
    # Calculate the coherence spectrum
    coh_periods = 1 / freq  # Periods corresponding to the frequency axis
    coherence = WCT

    #----------------------------------------------------------------------
    # Calculate significance levels using Monte Carlo randomization method
    #----------------------------------------------------------------------
    nxx, nyy = coherence.shape
    coh_perm = np.zeros((nxx, nyy, params['nperm']))
    print('\nCalculating wavelet coherence significance:')
    total_iterations = params['nperm']  # Total number of permutations
    iteration_count = 0  # To keep track of progress
    for ip in range(params['nperm']):
        iteration_count += 1
        progress = (iteration_count / total_iterations) * 100
        print(f"\rProgress: {progress:.2f}%", end="")
        
        y_perm1 = np.random.permutation(data1orig)  # Permuting the original signal
        apocube1 = WaLSA_detrend_apod(
            y_perm1, 
            apod=params['apod'], 
            meandetrend=params['meandetrend'], 
            pxdetrend=params['pxdetrend'], 
            polyfit=params['polyfit'], 
            meantemporal=params['meantemporal'], 
            recon=params['recon'], 
            cadence=cadence, 
            resample_original=params['resample_original'], 
            silent=True
        )
        y_perm2 = np.random.permutation(data2orig)  # Permuting the original signal
        apocube2 = WaLSA_detrend_apod(
            y_perm2, 
            apod=params['apod'], 
            meandetrend=params['meandetrend'], 
            pxdetrend=params['pxdetrend'], 
            polyfit=params['polyfit'], 
            meantemporal=params['meantemporal'], 
            recon=params['recon'], 
            cadence=cadence, 
            resample_original=params['resample_original'], 
            silent=True
        )
        WCTs, _, _, _, _ = wct(
            apocube1, apocube2, cadence, dj=params['dj'], s0=params['s0'], J=params['J'], 
            sig=False, wavelet=params['mother'], normalize=False, cache=params['cache']
        )
        coh_perm[:, :, ip] = WCTs

    sig_coh = WaLSA_wavelet_confidence(coh_perm, siglevel=params['siglevel'])

    coh_sig = coherence / sig_coh  # Ratio > 1 means coherence is significant

    # --------------- PHASE ANGLES ---------------
    phase_angle = aWCT

    return (
        cross_power, cross_periods, cross_sig, cross_coi, 
        coherence, coh_periods, coh_sig, coh_coi, 
        phase_angle
    )



#!/usr/bin/python
# coding: UTF-8
#
# Author:   Dawid Laszuk
# Issues:  https://github.com/laszukdawid/PyEMD/issues
"""
Uses Numba (https://numba.pydata.org/) as a Just-in-Time (JIT)
compiler for Python, mostly Numpy. Just-in-time compilation means that
the code is compiled (machine code) during execution, and thus shows
benefit when there's plenty of repeated code or same code used a lot.

This EMD implementation is experimental as it only provides value
when there's significant amount of computation required, e.g. when
analyzing HUGE time series with a lot of internal complexity,
or reuses the instance/method many times, e.g. in a script,
iPython REPL or jupyter notebook.

Additional reason for this being experimental is that the author (me)
isn't well veristile in Numba optimization. There's definitely a lot
that can be improved. It's being added as maybe it'll be helpful or
an inspiration for others to learn something and contribute to the PyEMD.

"""

import logging
from typing import Optional, Tuple

import numba as nb
import numpy as np
from numba.types import float64, int64, unicode_type
from scipy.interpolate import Akima1DInterpolator, interp1d

# from .utils import unify_types

FindExtremaOutput = Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]


default_emd_config = nb.typed.Dict.empty(
    key_type=unicode_type,
    value_type=float64,
)
default_emd_config["nbsym"] = 2.0
default_emd_config["energy_ratio_thr"] = 0.2
default_emd_config["std_thr"] = 0.2
default_emd_config["svar_thr"] = 0.001
default_emd_config["total_power_thr"] = 0.005
default_emd_config["range_thr"] = 0.001
default_emd_config["FIXE"] = 0.0
default_emd_config["FIXE_H"] = 0.0
default_emd_config["MAX_ITERATION"] = 1000.0


class JitEMD:
    def __init__(self, config=None, spline_kind="cubic", extrema_detection="simple"):
        self.config = config or default_emd_config
        self.spline_kind = spline_kind
        self.extrema_detection = extrema_detection
        self.imfs = None
        self.imfs = None
        self.residue = None

    def get_imfs_and_trend(self):
        if np.allclose(self.residue, 0):
            return self.imfs[:-1].copy(), self.imfs[-1].copy()
        return self.imfs, self.residue

    def emd(self, s, t, max_imf=-1):
        imfs = emd(
            s,
            t,
            max_imf=max_imf,
            spline_kind=self.spline_kind,
            extrema_detection=self.extrema_detection,
            config=self.config,
        )
        self.imfs = imfs[:-1, :]
        self.residue = imfs[-1:, :]
        return imfs

    def __call__(self, s, t, max_imf=-1):
        return self.emd(s, t, max_imf=max_imf)


@nb.jit(float64[:](float64[:], int64, float64[:]), nopython=True)
def np_round(x, decimals, out):
    return np.round_(x, decimals, out)


@nb.njit
def _not_duplicate(S: np.ndarray) -> np.ndarray:
    dup = np.logical_and(S[1:-1] == S[0:-2], S[1:-1] == S[2:])
    not_dup_idx = np.arange(1, len(S) - 1)[~dup]

    idx = np.empty(len(not_dup_idx) + 2, dtype=np.int64)
    idx[0] = 0
    idx[-1] = len(S) - 1
    idx[1:-1] = not_dup_idx

    return idx


@nb.jit
def cubic_spline_3pts(x, y, T):
    """
    Apparently scipy.interpolate.interp1d does not support
    cubic spline for less than 4 points.
    """
    x0, x1, x2 = x
    y0, y1, y2 = y

    x1x0, x2x1 = x1 - x0, x2 - x1
    y1y0, y2y1 = y1 - y0, y2 - y1
    _x1x0, _x2x1 = 1.0 / x1x0, 1.0 / x2x1

    m11, m12, m13 = 2 * _x1x0, _x1x0, 0
    m21, m22, m23 = _x1x0, 2.0 * (_x1x0 + _x2x1), _x2x1
    m31, m32, m33 = 0, _x2x1, 2.0 * _x2x1

    v1 = 3 * y1y0 * _x1x0 * _x1x0
    v3 = 3 * y2y1 * _x2x1 * _x2x1
    v2 = v1 + v3

    M = np.array([[m11, m12, m13], [m21, m22, m23], [m31, m32, m33]])
    v = np.array([v1, v2, v3]).T
    k = np.linalg.inv(M).dot(v)

    a1 = k[0] * x1x0 - y1y0
    b1 = -k[1] * x1x0 + y1y0
    a2 = k[1] * x2x1 - y2y1
    b2 = -k[2] * x2x1 + y2y1

    t = T[np.logical_and(T >= x0, T <= x2)]
    t1 = (T[np.logical_and(T >= x0, T < x1)] - x0) / x1x0
    t2 = (T[np.logical_and(T >= x1, T <= x2)] - x1) / x2x1
    t11, t22 = 1.0 - t1, 1.0 - t2

    q1 = t11 * y0 + t1 * y1 + t1 * t11 * (a1 * t11 + b1 * t1)
    q2 = t22 * y1 + t2 * y2 + t2 * t22 * (a2 * t22 + b2 * t2)
    q = np.append(q1, q2)

    return t, q


@nb.jit(nopython=False, forceobj=True)
def akima(X, Y, x):
    spl = Akima1DInterpolator(X, Y)
    return spl(x)


@nb.njit
def nb_diff(s):
    return s[1:] - s[:-1]


@nb.jit(
    nb.types.UniTuple(float64[:], 5)(float64[:], float64[:]),
    nopython=True,
)
def _find_extrema_simple(T: np.ndarray, S: np.ndarray) -> FindExtremaOutput:
    # Finds indexes of zero-crossings
    S1, S2 = S[:-1], S[1:]
    indzer = np.nonzero(S1 * S2 < 0)[0]
    if np.any(S == 0):
        indz = np.nonzero(S == 0)[0]
        if np.any(nb_diff(indz) == 1):
            zer = S == 0
            dz = nb_diff(np.append(np.append(0, zer), 0))
            debz = np.nonzero(dz == 1)[0].astype(np.float64)
            finz = (np.nonzero(dz == -1)[0] - 1).astype(np.float64)
            # indz = np.round((debz + finz) / 2)
            y = np.empty_like(debz)
            np_round((debz + finz) / 2, 0, y)
            indz = y.astype(np.int64)

        indzer = np.sort(np.append(indzer, indz))

    # Finds local extrema
    d = S2 - S1
    d1, d2 = d[:-1], d[1:]
    indmin = np.nonzero(np.logical_and(d1 * d2 < 0, d1 < 0))[0] + 1
    indmax = np.nonzero(np.logical_and(d1 * d2 < 0, d1 > 0))[0] + 1
    indmin = indmin.astype(np.int64)
    indmax = indmax.astype(np.int64)

    # When two or more points have the same value
    if np.any(d == 0):
        imax, imin = [], []

        same_values = d == 0
        dd = nb_diff(np.append(np.append(0, same_values), 0))
        _left_idx = np.nonzero(dd == 1)[0]
        _right_idx = np.nonzero(dd == -1)[0]
        if _left_idx[0] == 1:
            _left_idx = _left_idx[1:]
            _right_idx = _right_idx[1:]

        if len(_left_idx) > 0:
            if _right_idx[-1] == len(S) - 1:
                _left_idx = _left_idx[:-1]
                _right_idx = _right_idx[:-1]

        for k in range(len(_left_idx)):
            _left_value = d[_left_idx[k] - 1]
            _right_value = d[_right_idx[k]]
            _mid_value = round((_left_idx[k] + _right_idx[k]) / 2.0)
            if _left_value > 0 and _right_value < 0:
                imax.append(_mid_value)
            elif _left_value < 0 and _right_value > 0:
                imin.append(_mid_value)

        if len(imax) > 0:
            indmax = np.append(indmax, np.array(imax)).astype(np.int64)
            indmax.sort()

        if len(imin) > 0:
            indmin = np.append(indmin, np.array(imin)).astype(np.int64)
            indmin.sort()

    local_max_pos = T[indmax].astype(S.dtype)
    local_max_val = S[indmax].astype(S.dtype)
    local_min_pos = T[indmin].astype(S.dtype)
    local_min_val = S[indmin].astype(S.dtype)

    return local_max_pos, local_max_val, local_min_pos, local_min_val, indzer.astype(S.dtype)


@nb.jit(
    nb.types.Tuple((float64[:], float64[:], float64[:], float64[:], float64[:]))(float64[:], float64[:]),
    nopython=True,
)
def _find_extrema_parabol(T: np.ndarray, S: np.ndarray) -> FindExtremaOutput:
    # Finds indexes of zero-crossings
    S1, S2 = S[:-1], S[1:]
    indzer = np.nonzero(S1 * S2 < 0)[0]
    if np.any(S == 0):
        indz = np.nonzero(S == 0)[0]
        if np.any(nb_diff(indz) == 1):
            zer = S == 0
            dz = nb_diff(np.append(np.append(0, zer), 0))
            debz = (np.nonzero(dz == 1)[0]).astype(np.float64)
            finz = (np.nonzero(dz == -1)[0] - 1).astype(np.float64)
            y = np.empty_like(debz)
            np_round((debz + finz) / 2, 0, y)
            indz = y.astype(np.int64)

        indzer = np.sort(np.append(indzer, indz))

    dt = float(T[1] - T[0])
    scale = 2.0 * dt * dt

    idx = _not_duplicate(S)
    T = T[idx]
    S = S[idx]

    # p - previous
    # 0 - current
    # n - next
    Tp, T0, Tn = T[:-2], T[1:-1], T[2:]
    Sp, S0, Sn = S[:-2], S[1:-1], S[2:]
    # a = Sn + Sp - 2*S0
    # b = 2*(Tn+Tp)*S0 - ((Tn+T0)*Sp+(T0+Tp)*Sn)
    # c = Sp*T0*Tn -2*Tp*S0*Tn + Tp*T0*Sn
    TnTp, T0Tn, TpT0 = Tn - Tp, T0 - Tn, Tp - T0
    scale = Tp * Tn * Tn + Tp * Tp * T0 + T0 * T0 * Tn - Tp * Tp * Tn - Tp * T0 * T0 - T0 * Tn * Tn

    a = T0Tn * Sp + TnTp * S0 + TpT0 * Sn
    b = (S0 - Sn) * Tp**2 + (Sn - Sp) * T0**2 + (Sp - S0) * Tn**2
    c = T0 * Tn * T0Tn * Sp + Tn * Tp * TnTp * S0 + Tp * T0 * TpT0 * Sn

    a = a / scale
    b = b / scale
    c = c / scale
    a[a == 0] = 1e-14
    tVertex = -0.5 * b / a
    idx = np.logical_and(tVertex < T0 + 0.5 * (Tn - T0), tVertex >= T0 - 0.5 * (T0 - Tp))

    a, b, c = a[idx], b[idx], c[idx]
    tVertex = tVertex[idx]
    sVertex = a * tVertex * tVertex + b * tVertex + c

    local_max_pos, local_max_val = tVertex[a < 0], sVertex[a < 0]
    local_min_pos, local_min_val = tVertex[a > 0], sVertex[a > 0]

    return local_max_pos, local_max_val, local_min_pos, local_min_val, indzer.astype(local_max_val.dtype)


@nb.jit(
    nb.types.Tuple((float64[:], float64[:], float64[:], float64[:], float64[:]))(float64[:], float64[:], unicode_type),
    nopython=True,
)
def find_extrema(T: np.ndarray, S: np.ndarray, extrema_detection: str) -> FindExtremaOutput:
    """
    Returns extrema (minima and maxima) for given signal S.
    Detection and definition of the extrema depends on
    ``extrema_detection`` variable, set on initiation of EMD.

    Parameters
    ----------
    T : numpy array
        Position or time array.
    S : numpy array
        Input data S(T).

    Returns
    -------
    local_max_pos : numpy array
        Position of local maxima.
    local_max_val : numpy array
        Values of local maxima.
    local_min_pos : numpy array
        Position of local minima.
    local_min_val : numpy array
        Values of local minima.
    """
    # return _find_extrema_simple(T, S)
    if extrema_detection == "parabol":
        return _find_extrema_parabol(T, S)
    elif extrema_detection == "simple":
        return _find_extrema_simple(T, S)
    else:
        raise ValueError("Incorrect extrema detection type. Please try: 'simple' or 'parabol'.")


@nb.jit(nopython=True)
def prepare_points(
    T: np.ndarray,
    S: np.ndarray,
    max_pos: np.ndarray,
    max_val: np.ndarray,
    min_pos: np.ndarray,
    min_val: np.ndarray,
    extrema_detection,
    nbsym,
):
    """
    Performs extrapolation on edges by adding extra extrema, also known
    as mirroring signal. The number of added points depends on *nbsym*
    variable.

    Parameters
    ----------
    T : numpy array
        Position or time array.
    S : numpy array
        Input signal.
    max_pos : iterable
        Sorted time positions of maxima.
    max_val : iterable
        Signal values at max_pos positions.
    min_pos : iterable
        Sorted time positions of minima.
    min_val : iterable
        Signal values at min_pos positions.

    Returns
    -------
    max_extrema : numpy array (2 rows)
        Position (1st row) and values (2nd row) of minima.
    min_extrema : numpy array (2 rows)
        Position (1st row) and values (2nd row) of maxima.
    """
    if extrema_detection == "parabol":
        return _prepare_points_parabol(T, S, max_pos, max_val, min_pos, min_val, nbsym)
    elif extrema_detection == "simple":
        return _prepare_points_simple(T, S, max_pos, min_pos, nbsym)
    else:
        msg = "Incorrect extrema detection type. Please try: 'simple' or 'parabol'."
        raise ValueError(msg)


@nb.jit(nopython=True)
def _prepare_points_parabol(T, S, max_pos, max_val, min_pos, min_val, nbsym) -> Tuple[np.ndarray, np.ndarray]:
    # Need at least two extrema to perform mirroring
    DTYPE = S.dtype
    max_extrema = np.zeros((2, len(max_pos)), dtype=DTYPE)
    min_extrema = np.zeros((2, len(min_pos)), dtype=DTYPE)

    max_extrema[0], min_extrema[0] = max_pos, min_pos
    max_extrema[1], min_extrema[1] = max_val, min_val

    # Local variables
    end_min, end_max = len(min_pos), len(max_pos)

    ####################################
    # Left bound
    d_pos = max_pos[0] - min_pos[0]
    left_ext_max_type = d_pos < 0  # True -> max, else min

    # Left extremum is maximum
    if left_ext_max_type:
        if (S[0] > min_val[0]) and (np.abs(d_pos) > (max_pos[0] - T[0])):
            # mirror signal to first extrema
            expand_left_max_pos = 2 * max_pos[0] - max_pos[1 : nbsym + 1]
            expand_left_min_pos = 2 * max_pos[0] - min_pos[0:nbsym]
            expand_left_max_val = max_val[1 : nbsym + 1]
            expand_left_min_val = min_val[0:nbsym]
        else:
            # mirror signal to beginning
            expand_left_max_pos = 2 * T[0] - max_pos[0:nbsym]
            expand_left_min_pos = 2 * T[0] - np.append(T[0], min_pos[0 : nbsym - 1])
            expand_left_max_val = max_val[0:nbsym]
            expand_left_min_val = np.append(S[0], min_val[0 : nbsym - 1])

    # Left extremum is minimum
    else:
        if (S[0] < max_val[0]) and (np.abs(d_pos) > (min_pos[0] - T[0])):
            # mirror signal to first extrema
            expand_left_max_pos = 2 * min_pos[0] - max_pos[0:nbsym]
            expand_left_min_pos = 2 * min_pos[0] - min_pos[1 : nbsym + 1]
            expand_left_max_val = max_val[0:nbsym]
            expand_left_min_val = min_val[1 : nbsym + 1]
        else:
            # mirror signal to beginning
            expand_left_max_pos = 2 * T[0] - np.append(T[0], max_pos[0 : nbsym - 1])
            expand_left_min_pos = 2 * T[0] - min_pos[0:nbsym]
            expand_left_max_val = np.append(S[0], max_val[0 : nbsym - 1])
            expand_left_min_val = min_val[0:nbsym]

    if not expand_left_min_pos.shape:
        expand_left_min_pos, expand_left_min_val = min_pos, min_val
    if not expand_left_max_pos.shape:
        expand_left_max_pos, expand_left_max_val = max_pos, max_val

    expand_left_min = np.vstack((expand_left_min_pos[::-1], expand_left_min_val[::-1]))
    expand_left_max = np.vstack((expand_left_max_pos[::-1], expand_left_max_val[::-1]))

    ####################################
    # Right bound
    d_pos = max_pos[-1] - min_pos[-1]
    right_ext_max_type = d_pos > 0

    # Right extremum is maximum
    if not right_ext_max_type:
        if (S[-1] < max_val[-1]) and (np.abs(d_pos) > (T[-1] - min_pos[-1])):
            # mirror signal to last extrema
            idx_max = max(0, end_max - nbsym)
            idx_min = max(0, end_min - nbsym - 1)
            expand_right_max_pos = 2 * min_pos[-1] - max_pos[idx_max:]
            expand_right_min_pos = 2 * min_pos[-1] - min_pos[idx_min:-1]
            expand_right_max_val = max_val[idx_max:]
            expand_right_min_val = min_val[idx_min:-1]
        else:
            # mirror signal to end
            idx_max = max(0, end_max - nbsym + 1)
            idx_min = max(0, end_min - nbsym)
            expand_right_max_pos = 2 * T[-1] - np.append(max_pos[idx_max:], T[-1])
            expand_right_min_pos = 2 * T[-1] - min_pos[idx_min:]
            expand_right_max_val = np.append(max_val[idx_max:], S[-1])
            expand_right_min_val = min_val[idx_min:]

    # Right extremum is minimum
    else:
        if (S[-1] > min_val[-1]) and len(max_pos) > 1 and (np.abs(d_pos) > (T[-1] - max_pos[-1])):
            # mirror signal to last extremum
            idx_max = max(0, end_max - nbsym - 1)
            idx_min = max(0, end_min - nbsym)
            expand_right_max_pos = 2 * max_pos[-1] - max_pos[idx_max:-1]
            expand_right_min_pos = 2 * max_pos[-1] - min_pos[idx_min:]
            expand_right_max_val = max_val[idx_max:-1]
            expand_right_min_val = min_val[idx_min:]
        else:
            # mirror signal to end
            idx_max = max(0, end_max - nbsym)
            idx_min = max(0, end_min - nbsym + 1)
            expand_right_max_pos = 2 * T[-1] - max_pos[idx_max:]
            expand_right_min_pos = 2 * T[-1] - np.append(min_pos[idx_min:], T[-1])
            expand_right_max_val = max_val[idx_max:]
            expand_right_min_val = np.append(min_val[idx_min:], S[-1])

    if not expand_right_min_pos.shape:
        expand_right_min_pos, expand_right_min_val = min_pos, min_val
    if not expand_right_max_pos.shape:
        expand_right_max_pos, expand_right_max_val = max_pos, max_val

    expand_right_min = np.vstack((expand_right_min_pos[::-1], expand_right_min_val[::-1]))
    expand_right_max = np.vstack((expand_right_max_pos[::-1], expand_right_max_val[::-1]))

    max_extrema = np.hstack((expand_left_max, max_extrema, expand_right_max))
    min_extrema = np.hstack((expand_left_min, min_extrema, expand_right_min))

    return max_extrema, min_extrema


@nb.jit(nopython=True)
def _prepare_points_simple(
    T: np.ndarray,
    S: np.ndarray,
    max_pos: np.ndarray,
    min_pos: np.ndarray,
    nbsym: int,
) -> Tuple[np.ndarray, np.ndarray]:
    # Find indexes of pass
    ind_min = min_pos.astype(np.int64)
    ind_max = max_pos.astype(np.int64)

    # Local variables
    end_min, end_max = len(min_pos), len(max_pos)

    ####################################
    # Left bound - mirror nbsym points to the left
    if ind_max[0] < ind_min[0]:
        if S[0] > S[ind_min[0]]:
            lmax = ind_max[1 : min(end_max, nbsym + 1)][::-1]
            lmin = ind_min[0 : min(end_min, nbsym + 0)][::-1]
            lsym = ind_max[0]
        else:
            lmax = ind_max[0 : min(end_max, nbsym)][::-1]
            lmin = np.append(ind_min[0 : min(end_min, nbsym - 1)][::-1], 0)
            lsym = 0
    else:
        if S[0] < S[ind_max[0]]:
            lmax = ind_max[0 : min(end_max, nbsym + 0)][::-1]
            lmin = ind_min[1 : min(end_min, nbsym + 1)][::-1]
            lsym = ind_min[0]
        else:
            lmax = np.append(ind_max[0 : min(end_max, nbsym - 1)][::-1], 0)
            lmin = ind_min[0 : min(end_min, nbsym)][::-1]
            lsym = 0

    ####################################
    # Right bound - mirror nbsym points to the right
    if ind_max[-1] < ind_min[-1]:
        if S[-1] < S[ind_max[-1]]:
            rmax = ind_max[max(end_max - nbsym, 0) :][::-1]
            rmin = ind_min[max(end_min - nbsym - 1, 0) : -1][::-1]
            rsym = ind_min[-1]
        else:
            rmax = np.append(ind_max[max(end_max - nbsym + 1, 0) :], len(S) - 1)[::-1]
            rmin = ind_min[max(end_min - nbsym, 0) :][::-1]
            rsym = len(S) - 1
    else:
        if S[-1] > S[ind_min[-1]]:
            rmax = ind_max[max(end_max - nbsym - 1, 0) : -1][::-1]
            rmin = ind_min[max(end_min - nbsym, 0) :][::-1]
            rsym = ind_max[-1]
        else:
            rmax = ind_max[max(end_max - nbsym, 0) :][::-1]
            rmin = np.append(ind_min[max(end_min - nbsym + 1, 0) :], len(S) - 1)[::-1]
            rsym = len(S) - 1

    # In case any array missing
    if not lmin.size:
        lmin = ind_min
    if not rmin.size:
        rmin = ind_min
    if not lmax.size:
        lmax = ind_max
    if not rmax.size:
        rmax = ind_max

    # Mirror points
    tlmin = 2 * T[lsym] - T[lmin]
    tlmax = 2 * T[lsym] - T[lmax]
    trmin = 2 * T[rsym] - T[rmin]
    trmax = 2 * T[rsym] - T[rmax]

    # If mirrored points are not outside passed time range.
    if tlmin[0] > T[0] or tlmax[0] > T[0]:
        if lsym == ind_max[0]:
            lmax = ind_max[0 : min(end_max, nbsym)][::-1]
        else:
            lmin = ind_min[0 : min(end_min, nbsym)][::-1]

        if lsym == 0:
            raise Exception("Left edge BUG")

        lsym = 0
        tlmin = 2 * T[lsym] - T[lmin]
        tlmax = 2 * T[lsym] - T[lmax]

    if trmin[-1] < T[-1] or trmax[-1] < T[-1]:
        if rsym == ind_max[-1]:
            rmax = ind_max[max(end_max - nbsym, 0) :][::-1]
        else:
            rmin = ind_min[max(end_min - nbsym, 0) :][::-1]

        if rsym == len(S) - 1:
            raise Exception("Right edge BUG")

        rsym = len(S) - 1
        trmin = 2 * T[rsym] - T[rmin]
        trmax = 2 * T[rsym] - T[rmax]

    zlmax = S[lmax]
    zlmin = S[lmin]
    zrmax = S[rmax]
    zrmin = S[rmin]

    tmin = np.append(tlmin, np.append(T[ind_min], trmin))
    tmax = np.append(tlmax, np.append(T[ind_max], trmax))
    zmin = np.append(zlmin, np.append(S[ind_min], zrmin))
    zmax = np.append(zlmax, np.append(S[ind_max], zrmax))

    max_extrema = np.vstack((tmax, zmax))
    min_extrema = np.vstack((tmin, zmin))

    # For posterity:
    #  I tried with np.delete and np.vstack([ ]) but both didn't work.
    #  np.delete works only with 2 args, and vstack had problem with list comphr.
    max_dup_idx = np.where(max_extrema[0, 1:] == max_extrema[0, :-1])[0]
    if len(max_dup_idx):
        for col_idx in max_dup_idx:
            max_extrema = np.hstack((max_extrema[:, :col_idx], max_extrema[:, col_idx + 1 :]))

    min_dup_idx = np.where(min_extrema[0, 1:] == min_extrema[0, :-1])[0]
    if len(min_dup_idx):
        for col_idx in min_dup_idx:
            min_extrema = np.hstack((min_extrema[:, :col_idx], min_extrema[:, col_idx + 1 :]))

    return max_extrema, min_extrema


@nb.jit(nopython=False, forceobj=True)
def spline_points(T: np.ndarray, extrema: np.ndarray, spline_kind: str) -> Tuple[np.ndarray, np.ndarray]:
    dtype = extrema.dtype
    kind = spline_kind.lower()
    t = T[np.logical_and(T >= extrema[0, 0], T <= extrema[0, -1])]

    if kind == "akima":
        return t, akima(extrema[0], extrema[1], t)

    elif kind == "cubic":
        if extrema.shape[1] > 3:
            interpolation = interp1d(extrema[0], extrema[1], kind=kind)
            interpolated: np.ndarray = interpolation(t)
            return t, interpolated.astype(dtype)
        else:
            return cubic_spline_3pts(extrema[0], extrema[1], t)

    elif kind in ["slinear", "quadratic", "linear"]:
        interpolation = interp1d(extrema[0], extrema[1], kind=kind)
        interpolated: np.ndarray = interpolation(t)
        return T, interpolated.astype(dtype)

    else:
        raise ValueError("No such interpolation method!")


@nb.jit(
    nb.types.UniTuple(float64[:, :], 2)(float64[:], float64[:], int64, unicode_type),
    nopython=True,
)
def extract_max_min_extrema(
    T: np.ndarray,
    S: np.ndarray,
    nbsym: int,
    extrema_detection: str,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extracts top and bottom envelopes based on the signal,
    which are constructed based on maxima and minima, respectively.

    Parameters
    ----------
    T : numpy array
        Position or time array.
    S : numpy array
        Input data S(T).

    Returns
    -------
    max_spline : numpy array
        Spline spanned on S maxima.
    min_spline : numpy array
        Spline spanned on S minima.
    max_extrema : numpy array
        Points indicating local maxima.
    min_extrema : numpy array
        Points indicating local minima.
    """

    # Get indexes of extrema
    ext_res = find_extrema(T, S, extrema_detection)
    max_pos, max_val = ext_res[0], ext_res[1]
    min_pos, min_val = ext_res[2], ext_res[3]

    if len(max_pos) + len(min_pos) < 3:
        minus_one = -1 * np.ones((1, 1), dtype=S.dtype)
        return minus_one.copy(), minus_one.copy()

    #########################################
    # Extrapolation of signal (over boundaries)
    max_extrema, min_extrema = prepare_points(T, S, max_pos, max_val, min_pos, min_val, extrema_detection, nbsym)
    return max_extrema, min_extrema


@nb.njit
def end_condition(S: np.ndarray, IMF: np.ndarray, range_thr: float, total_power_thr: float) -> bool:
    """Tests for end condition of whole EMD. The procedure will stop if:

    * Absolute amplitude (max - min) is below *range_thr* threshold, or
    * Metric L1 (mean absolute difference) is below *total_power_thr* threshold.

    Parameters
    ----------
    S : numpy array
        Original signal on which EMD was performed.
    IMF : numpy 2D array
        Set of IMFs where each row is IMF. Their order is not important.

    Returns
    -------
    end : bool
        Whether sifting is finished.
    """
    # When to stop EMD
    tmp = S - np.sum(IMF, axis=0)

    if np.max(tmp) - np.min(tmp) < range_thr:
        return True

    if np.sum(np.abs(tmp)) < total_power_thr:
        return True

    return False


@nb.jit
def check_imf(
    imf_new: np.ndarray,
    imf_old: np.ndarray,
    eMax: np.ndarray,
    eMin: np.ndarray,
    svar_thr: float,
    std_thr: float,
    energy_ratio_thr: float,
) -> bool:
    """
    Huang criteria for **IMF** (similar to Cauchy convergence test).
    Signal is an IMF if consecutive siftings do not affect signal
    in a significant manner.
    """
    # local max are >0 and local min are <0
    # if np.any(eMax[1] < 0) or np.any(eMin[1] > 0):
    if np.sum(eMax[1] < 0) + np.sum(eMin[1] > 0) > 0:
        return False

    # Convergence
    if np.sum(imf_new**2) < 1e-10:
        return False

    # Precompute values
    imf_diff = imf_new - imf_old
    imf_diff_sqrd_sum = np.sum(imf_diff * imf_diff)

    # Scaled variance test
    svar = imf_diff_sqrd_sum / (max(imf_old) - min(imf_old))
    if svar < svar_thr:
        return True

    # Standard deviation test
    std = np.sum((imf_diff / imf_new) ** 2)
    if std < std_thr:
        return True

    energy_ratio = imf_diff_sqrd_sum / np.sum(imf_old * imf_old)
    if energy_ratio < energy_ratio_thr:
        return True

    return False


@nb.jit
def _normalize_time(t: np.ndarray) -> np.ndarray:
    """
    Normalize time array so that it doesn't explode on tiny values.
    Returned array starts with 0 and the smallest increase is by 1.
    """
    d = nb_diff(t)
    assert np.all(d != 0), "All time domain values needs to be unique"
    return (t - t[0]) / np.min(d)


@nb.jit(
    float64[:, :](
        float64[:],
        float64[:],
        nb.optional(int64),
        nb.optional(unicode_type),
        nb.optional(unicode_type),
        nb.optional(nb.typeof(default_emd_config)),
    ),
    nopython=False,
    forceobj=True,
)
def emd(
    S: np.ndarray,
    T: np.ndarray,
    max_imf: int = -1,
    spline_kind: str = "cubic",
    extrema_detection: str = "simple",
    config=default_emd_config,
) -> np.ndarray:
    # if T is not None and len(S) != len(T):
    #     raise ValueError("Time series have different sizes: len(S) -> {} != {} <- len(T)".format(len(S), len(T)))
    #     return None

    # if T is None or config["extrema_detection"] == "simple":
    #     T = get_timeline(len(S), S.dtype)

    # Normalize T so that it doesn't explode
    # T = _normalize_time(T)

    # Make sure same types are dealt
    # S, T = unify_types(S, T)
    MAX_ITERATION = config["MAX_ITERATION"]
    FIXE = config["FIXE"]
    FIXE_H = config["FIXE_H"]
    nbsym = config["nbsym"]
    svar_thr, std_thr, energy_ratio_thr = config["svar_thr"], config["std_thr"], config["energy_ratio_thr"]
    range_thr, total_power_thr = config["range_thr"], config["total_power_thr"]

    DTYPE = S.dtype
    N = len(S)

    residue = S.astype(DTYPE)
    imf = np.zeros(len(S), dtype=DTYPE)
    imf_old = np.nan

    if S.shape != T.shape:
        raise ValueError("Position or time array should be the same size as signal.")

    # Create arrays
    imfNo = 0
    extNo = -1
    IMF = np.empty((imfNo, N))  # Numpy container for IMF
    finished = False

    while not finished:
        residue[:] = S - np.sum(IMF[:imfNo], axis=0)
        imf = residue.copy()
        mean = np.zeros(len(S), dtype=DTYPE)

        # Counters
        n = 0  # All iterations for current imf.
        n_h = 0  # counts when |#zero - #ext| <=1

        while True:
            n += 1
            if n >= MAX_ITERATION:
                break

            ext_res = find_extrema(T, imf, extrema_detection)
            max_pos, min_pos, indzer = ext_res[0], ext_res[2], ext_res[4]
            extNo = len(min_pos) + len(max_pos)
            nzm = len(indzer)

            if extNo > 2:
                max_extrema, min_extrema = extract_max_min_extrema(T, imf, nbsym, extrema_detection)
                _, max_env = spline_points(T, max_extrema, spline_kind)
                _, min_env = spline_points(T, min_extrema, spline_kind)
                mean = 0.5 * (max_env + min_env)

                imf_old = imf.copy()
                imf = imf - mean

                # Fix number of iterations
                if FIXE:
                    if n >= FIXE:
                        break

                # Fix number of iterations after number of zero-crossings
                # and extrema differ at most by one.
                elif FIXE_H:
                    max_pos, _, min_pos, _, ind_zer = find_extrema(T, imf, extrema_detection)
                    extNo = len(max_pos) + len(min_pos)
                    nzm = len(ind_zer)

                    if n == 1:
                        continue

                    # If proto-IMF add one, or reset counter otherwise
                    n_h = n_h + 1 if abs(extNo - nzm) < 2 else 0

                    # STOP
                    if n_h >= FIXE_H:
                        break

                # Stops after default stopping criteria are met
                else:
                    ext_res = find_extrema(T, imf, extrema_detection)
                    max_pos, _, min_pos, _, ind_zer = ext_res
                    extNo = len(max_pos) + len(min_pos)
                    nzm = len(ind_zer)

                    if imf_old is np.nan:
                        continue

                    f1 = check_imf(imf, imf_old, max_extrema, min_extrema, svar_thr, std_thr, energy_ratio_thr)
                    f2 = abs(extNo - nzm) < 2

                    # STOP
                    if f1 and f2:
                        break

            else:  # Less than 2 ext, i.e. trend
                finished = True
                break
        # END OF IMF SIFTING

        IMF = np.vstack((IMF, imf.copy()))
        imfNo += 1

        if end_condition(S, IMF, range_thr, total_power_thr) or imfNo == max_imf:
            finished = True
            break

    # If the last sifting had 2 or less extrema then that's a trend (residue)
    if extNo <= 2:
        IMF = IMF[:-1]

    # Saving imfs and residue for external references
    imfs = IMF.copy()
    residue = S - np.sum(imfs, axis=0)

    # If residue isn't 0 then add it to the output
    if not np.allclose(residue, 0):
        IMF = np.vstack((IMF, residue))

    return IMF


def get_timeline(range_max: int, dtype: Optional[np.dtype] = None) -> np.ndarray:
    """Returns timeline array for requirements.

    Parameters
    ----------
    range_max : int
        Largest value in range. Assume `range(range_max)`. Commonly that's length of the signal.
    dtype : np.dtype
        Minimal definition type. Returned timeline will have dtype that's the same or with higher byte size.

    """
    timeline = np.arange(0, range_max, dtype=dtype)
    if timeline[-1] != range_max - 1:
        inclusive_dtype = smallest_inclusive_dtype(timeline.dtype, range_max)
        timeline = np.arange(0, range_max, dtype=inclusive_dtype)
    return timeline


def smallest_inclusive_dtype(ref_dtype: np.dtype, ref_value) -> np.dtype:
    """Returns a numpy dtype with the same base as reference dtype (ref_dtype)
    but with the range that includes reference value (ref_value).

    Parameters
    ----------
    ref_dtype : dtype
         Reference dtype. Used to select the base, i.e. int or float, for returned type.
    ref_value : value
        A value which needs to be included in returned dtype. Value will be typically int or float.

    """
    # Integer path
    if np.issubdtype(ref_dtype, np.integer):
        for dtype in [np.uint16, np.uint32, np.uint64]:
            if ref_value < np.iinfo(dtype).max:
                return dtype
        max_val = np.iinfo(np.uint32).max
        raise ValueError("Requested too large integer range. Exceeds max( uint64 ) == '{}.".format(max_val))

    # Integer path
    if np.issubdtype(ref_dtype, np.floating):
        for dtype in [np.float16, np.float32, np.float64]:
            if ref_value < np.finfo(dtype).max:
                return dtype
        max_val = np.finfo(np.float64).max
        raise ValueError("Requested too large integer range. Exceeds max( float64 ) == '{}.".format(max_val))

    raise ValueError("Unsupported dtype '{}'. Only intX and floatX are supported.".format(ref_dtype))


###################################################


if __name__ == "__main__":
    import pylab as plt

    # Logging options
    logging.basicConfig(level=logging.DEBUG)

    # EMD options
    max_imf = -1
    DTYPE = np.float64

    # Signal options
    N = 400
    tMin, tMax = 0, 2 * np.pi
    T = np.linspace(tMin, tMax, N, dtype=DTYPE)

    S = np.sin(20 * T * (1 + 0.2 * T)) + T**2 + np.sin(13 * T)
    S = S.astype(DTYPE)
    print("Input S.dtype: " + str(S.dtype))

    # Prepare and run EMD
    config = default_emd_config
    imfs = emd(config, S, T, max_imf)
    imfNo = imfs.shape[0]

    # Plot results
    c = 1
    r = np.ceil((imfNo + 1) / c)

    plt.ioff()
    plt.subplot(r, c, 1)
    plt.plot(T, S, "r")
    plt.xlim((tMin, tMax))
    plt.title("Original signal")

    for num in range(imfNo):
        plt.subplot(r, c, num + 2)
        plt.plot(T, imfs[num], "g")
        plt.xlim((tMin, tMax))
        plt.ylabel("Imf " + str(num + 1))

    plt.tight_layout()
    plt.show()



# -----------------------------------------------------------------------------------------------------
# WaLSAtools: - Wave analysis tools
# Copyright (C) 2025 WaLSA Team - Shahin Jafarzadeh et al.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Note: If you use WaLSAtools for research, please consider citing:
# Jafarzadeh, S., Jess, D. B., Stangalini, M. et al. 2025, Nature Reviews Methods Primers, in press.
# -----------------------------------------------------------------------------------------------------

# The following code is a modified version of the PyCWT package.
# The original code and licence:
# PyCWT is released under the open source 3-Clause BSD license:
#
# Copyright (c) 2023 Sebastian Krieger, Nabil Freij, and contributors. All rights
# reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#----------------------------------------------------------------------------------
# Modifications by Shahin Jafarzadeh, 2024
#----------------------------------------------------------------------------------

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np # type: ignore
from tqdm import tqdm # type: ignore
from scipy.stats import chi2 # type: ignore

# Try to import the Python wrapper for FFTW.
try:
    import pyfftw.interfaces.scipy_fftpack as fft # type: ignore
    from multiprocessing import cpu_count

    # Fast planning, use all available threads.
    _FFTW_KWARGS_DEFAULT = {'planner_effort': 'FFTW_ESTIMATE',
                            'threads': cpu_count()}

    def fft_kwargs(signal, **kwargs):
        """Return optimized keyword arguments for FFTW"""
        kwargs.update(_FFTW_KWARGS_DEFAULT)
        kwargs['n'] = len(signal)  # do not pad
        return kwargs

# Otherwise, fall back to 2 ** n padded scipy FFTPACK
except ImportError:
    import scipy.fftpack as fft # type: ignore
    # Can be turned off, e.g. for MKL optimizations
    _FFT_NEXT_POW2 = True

    def fft_kwargs(signal, **kwargs):
        """Return next higher power of 2 for given signal to speed up FFT"""
        if _FFT_NEXT_POW2:
            return {'n': int(2 ** np.ceil(np.log2(len(signal))))}

from scipy.signal import lfilter # type: ignore
from os import makedirs
from os.path import exists, expanduser


def find(condition):
    """Returns the indices where ravel(condition) is true."""
    res, = np.nonzero(np.ravel(condition))
    return res


def ar1(x):
    """
    Allen and Smith autoregressive lag-1 autocorrelation coefficient.
    In an AR(1) model

        x(t) - <x> = gamma(x(t-1) - <x>) + \alpha z(t) ,

    where <x> is the process mean, gamma and \alpha are process
    parameters and z(t) is a Gaussian unit-variance white noise.

    Parameters
    ----------
    x : numpy.ndarray, list
        Univariate time series

    Returns
    -------
    g : float
        Estimate of the lag-one autocorrelation.
    a : float
        Estimate of the noise variance [var(x) ~= a**2/(1-g**2)]
    mu2 : float
        Estimated square on the mean of a finite segment of AR(1)
        noise, mormalized by the process variance.

    References
    ----------
    [1] Allen, M. R. and Smith, L. A. Monte Carlo SSA: detecting
        irregular oscillations in the presence of colored noise.
        *Journal of Climate*, **1996**, 9(12), 3373-3404.
        <http://dx.doi.org/10.1175/1520-0442(1996)009<3373:MCSDIO>2.0.CO;2>
    [2] http://www.madsci.org/posts/archives/may97/864012045.Eg.r.html

    """
    x = np.asarray(x)
    N = x.size
    xm = x.mean()
    x = x - xm

    # Estimates the lag zero and one covariance
    c0 = x.transpose().dot(x) / N
    c1 = x[0:N-1].transpose().dot(x[1:N]) / (N - 1)

    # According to A. Grinsteds' substitutions
    B = -c1 * N - c0 * N**2 - 2 * c0 + 2 * c1 - c1 * N**2 + c0 * N
    A = c0 * N**2
    C = N * (c0 + c1 * N - c1)
    D = B**2 - 4 * A * C

    if D > 0:
        g = (-B - D**0.5) / (2 * A)
    else:
        raise Warning('Cannot place an upperbound on the unbiased AR(1). '
                      'Series is too short or trend is to large.')

    # According to Allen & Smith (1996), footnote 4
    mu2 = -1 / N + (2 / N**2) * ((N - g**N) / (1 - g) -
                                 g * (1 - g**(N - 1)) / (1 - g)**2)
    c0t = c0 / (1 - mu2)
    a = ((1 - g**2) * c0t) ** 0.5

    return g, a, mu2


def ar1_spectrum(freqs, ar1=0.):
    """
    Lag-1 autoregressive theoretical power spectrum.

    Parameters
    ----------
    freqs : numpy.ndarray, list
        Frequencies at which to calculate the theoretical power
        spectrum.
    ar1 : float
        Autoregressive lag-1 correlation coefficient.

    Returns
    -------
    Pk : numpy.ndarray
        Theoretical discrete Fourier power spectrum of noise signal.

    References
    ----------
    [1] http://www.madsci.org/posts/archives/may97/864012045.Eg.r.html

    """
    # According to a post from the MadSci Network available at
    # http://www.madsci.org/posts/archives/may97/864012045.Eg.r.html,
    # the time-series spectrum for an auto-regressive model can be
    # represented as
    #
    # P_k = \frac{E}{\left|1- \sum\limits_{k=1}^{K} a_k \, e^{2 i \pi
    #   \frac{k f}{f_s} } \right|^2}
    #
    # which for an AR1 model reduces to
    #
    freqs = np.asarray(freqs)
    Pk = (1 - ar1 ** 2) / np.abs(1 - ar1 * np.exp(-2 * np.pi * 1j * freqs)) \
        ** 2

    return Pk


def rednoise(N, g, a=1.):
    """
    Red noise generator using filter.

    Parameters
    ----------
    N : int
        Length of the desired time series.
    g : float
        Lag-1 autocorrelation coefficient.
    a : float, optional
        Noise innovation variance parameter.

    Returns
    -------
    y : numpy.ndarray
        Red noise time series.

    """
    if g == 0:
        yr = np.randn(N, 1) * a
    else:
        # Twice the decorrelation time.
        tau = int(np.ceil(-2 / np.log(np.abs(g))))
        yr = lfilter([1, 0], [1, -g], np.random.randn(N + tau, 1) * a)
        yr = yr[tau:]

    return yr.flatten()


def rect(x, normalize=False):
    """TODO: describe what I do."""
    if type(x) in [int, float]:
        shape = [x, ]
    elif type(x) in [list, dict]:
        shape = x
    elif type(x) in [np.ndarray, np.ma.core.MaskedArray]:
        shape = x.shape
    X = np.zeros(shape)
    X[0] = X[-1] = 0.5
    X[1:-1] = 1

    if normalize:
        X /= X.sum()

    return X


def boxpdf(x):
    """
    Forces the probability density function of the input data to have
    a boxed distribution.

    Parameters
    ----------
    x (array like) :
        Input data

    Returns
    -------
    X (array like) :
        Boxed data varying between zero and one.
    Bx, By (array like) :
        Data lookup table.

    """
    import numpy as np
    x = np.asarray(x)
    n = x.size

    # Kind of 'unique'
    i = np.argsort(x)
    d = (np.diff(x[i]) != 0)
    j = find(np.concatenate([d, [True]]))
    X = x[i][j]

    j = np.concatenate([[0], j + 1])
    Y = 0.5 * (j[0:-1] + j[1:]) / n
    bX = np.interp(x, X, Y)

    return bX, X, Y


def get_cache_dir():
    """Returns the location of the cache directory."""
    # Sets cache directory according to user home path.
    cache_dir = '{}/.cache/pycwt/'.format(expanduser('~'))
    # Creates cache directory if not existant.
    if not exists(cache_dir):
        makedirs(cache_dir)
    # Returns cache directory.
    return cache_dir


import numpy as np
from scipy.special import gamma
from scipy.signal import convolve2d
from scipy.special.orthogonal import hermitenorm


class Morlet(object):
    """Implements the Morlet wavelet class.

    Note that the input parameters f and f0 are angular frequencies.
    f0 should be more than 0.8 for this function to be correct, its
    default value is f0 = 6.

    """

    def __init__(self, f0=6):
        self._set_f0(f0)
        self.name = 'Morlet'

    def psi_ft(self, f):
        """Fourier transform of the approximate Morlet wavelet."""
        return (np.pi ** -0.25) * np.exp(-0.5 * (f - self.f0) ** 2)

    def psi(self, t):
        """Morlet wavelet as described in Torrence and Compo (1998)."""
        return (np.pi ** -0.25) * np.exp(1j * self.f0 * t - t ** 2 / 2)

    def flambda(self):
        """Fourier wavelength as of Torrence and Compo (1998)."""
        return (4 * np.pi) / (self.f0 + np.sqrt(2 + self.f0 ** 2))

    def coi(self):
        """e-Folding Time as of Torrence and Compo (1998)."""
        return 1. / np.sqrt(2)

    def sup(self):
        """Wavelet support defined by the e-Folding time."""
        return 1. / self.coi

    def _set_f0(self, f0):
        # Sets the Morlet wave number, the degrees of freedom and the
        # empirically derived factors for the wavelet bases C_{\delta},
        # gamma, \delta j_0 (Torrence and Compo, 1998, Table 2)
        self.f0 = f0             # Wave number
        self.dofmin = 2          # Minimum degrees of freedom
        if self.f0 == 6:
            self.cdelta = 0.776  # Reconstruction factor
            self.gamma = 2.32    # Decorrelation factor for time averaging
            self.deltaj0 = 0.60  # Factor for scale averaging
        else:
            self.cdelta = -1
            self.gamma = -1
            self.deltaj0 = -1

    def smooth(self, W, dt, dj, scales):
        """Smoothing function used in coherence analysis.

        Parameters
        ----------
        W :
        dt :
        dj :
        scales :

        Returns
        -------
        T :

        """
        # The smoothing is performed by using a filter given by the absolute
        # value of the wavelet function at each scale, normalized to have a
        # total weight of unity, according to suggestions by Torrence &
        # Webster (1999) and by Grinsted et al. (2004).
        m, n = W.shape

        # Filter in time.
        k = 2 * np.pi * fft.fftfreq(fft_kwargs(W[0, :])['n'])
        k2 = k ** 2
        snorm = scales / dt
        # Smoothing by Gaussian window (absolute value of wavelet function)
        # using the convolution theorem: multiplication by Gaussian curve in
        # Fourier domain for each scale, outer product of scale and frequency
        F = np.exp(-0.5 * (snorm[:, np.newaxis] ** 2) * k2)  # Outer product
        smooth = fft.ifft(F * fft.fft(W, axis=1, **fft_kwargs(W[0, :])),
                          axis=1,  # Along Fourier frequencies
                          **fft_kwargs(W[0, :], overwrite_x=True))
        T = smooth[:, :n]  # Remove possibly padded region due to FFT

        if np.isreal(W).all():
            T = T.real

        # Filter in scale. For the Morlet wavelet it's simply a boxcar with
        # 0.6 width.
        wsize = self.deltaj0 / dj * 2
        win = rect(int(np.round(wsize)), normalize=True)
        T = convolve2d(T, win[:, np.newaxis], 'same')  # Scales are "vertical"

        return T


class Paul(object):
    """Implements the Paul wavelet class.

    Note that the input parameter f is the angular frequency and that
    the default order for this wavelet is m=4.

    """
    def __init__(self, m=4):
        self._set_m(m)
        self.name = 'Paul'

    # def psi_ft(self, f):
    #     """Fourier transform of the Paul wavelet."""
    #     return (2 ** self.m /
    #             np.sqrt(self.m * np.prod(range(2, 2 * self.m))) *
    #             f ** self.m * np.exp(-f) * (f > 0))
    
    def psi_ft(self, f): # modified by SJ
        """Fourier transform of the Paul wavelet with limits to prevent underflow."""
        expnt = -f
        expnt[expnt < -100] = -100  # Apply the threshold to avoid extreme values
        return (2 ** self.m /
                np.sqrt(self.m * np.prod(range(2, 2 * self.m))) *
                f ** self.m * np.exp(expnt) * (f > 0))

    def psi(self, t):
        """Paul wavelet as described in Torrence and Compo (1998)."""
        return (2 ** self.m * 1j ** self.m * np.prod(range(2, self.m - 1)) /
                np.sqrt(np.pi * np.prod(range(2, 2 * self.m + 1))) *
                (1 - 1j * t) ** (-(self.m + 1)))

    def flambda(self):
        """Fourier wavelength as of Torrence and Compo (1998)."""
        return 4 * np.pi / (2 * self.m + 1)

    def coi(self):
        """e-Folding Time as of Torrence and Compo (1998)."""
        return np.sqrt(2)

    def sup(self):
        """Wavelet support defined by the e-Folding time."""
        return 1 / self.coi

    def _set_m(self, m):
        # Sets the m derivative of a Gaussian, the degrees of freedom and the
        # empirically derived factors for the wavelet bases C_{\delta},
        # gamma, \delta j_0 (Torrence and Compo, 1998, Table 2)
        self.m = m               # Wavelet order
        self.dofmin = 2          # Minimum degrees of freedom
        if self.m == 4:
            self.cdelta = 1.132  # Reconstruction factor
            self.gamma = 1.17    # Decorrelation factor for time averaging
            self.deltaj0 = 1.50  # Factor for scale averaging
        else:
            self.cdelta = -1
            self.gamma = -1
            self.deltaj0 = -1


class DOG(object):
    """Implements the derivative of a Guassian wavelet class.

    Note that the input parameter f is the angular frequency and that
    for m=2 the DOG becomes the Mexican hat wavelet, which is then
    default.

    """
    def __init__(self, m=2):
        self._set_m(m)
        self.name = 'DOG'

    def psi_ft(self, f):
        """Fourier transform of the DOG wavelet."""
        return (- 1j ** self.m / np.sqrt(gamma(self.m + 0.5)) * f ** self.m *
                np.exp(- 0.5 * f ** 2))

    def psi(self, t):
        """DOG wavelet as described in Torrence and Compo (1998).

        The derivative of a Gaussian of order `n` can be determined using
        the probabilistic Hermite polynomials. They are explicitly
        written as:
            Hn(x) = 2 ** (-n / s) * n! * sum ((-1) ** m) *
                    (2 ** 0.5 * x) ** (n - 2 * m) / (m! * (n - 2*m)!)
        or in the recursive form:
            Hn(x) = x * Hn(x) - nHn-1(x)

        Source: http://www.ask.com/wiki/Hermite_polynomials

        """
        p = hermitenorm(self.m)
        return ((-1) ** (self.m + 1) * np.polyval(p, t) *
                np.exp(-t ** 2 / 2) / np.sqrt(gamma(self.m + 0.5)))

    def flambda(self):
        """Fourier wavelength as of Torrence and Compo (1998)."""
        return (2 * np.pi / np.sqrt(self.m + 0.5))

    def coi(self):
        """e-Folding Time as of Torrence and Compo (1998)."""
        return 1 / np.sqrt(2)

    def sup(self):
        """Wavelet support defined by the e-Folding time."""
        return 1 / self.coi

    def _set_m(self, m):
        # Sets the m derivative of a Gaussian, the degrees of freedom and the
        # empirically derived factors for the wavelet bases C_{\delta},
        # gamma, \delta j_0 (Torrence and Compo, 1998, Table 2).
        self.m = m               # m-derivative
        self.dofmin = 1          # Minimum degrees of freedom
        if self.m == 2:
            self.cdelta = 3.541  # Reconstruction factor
            self.gamma = 1.43    # Decorrelation factor for time averaging
            self.deltaj0 = 1.40  # Factor for scale averaging
        elif self.m == 6:
            self.cdelta = 1.966
            self.gamma = 1.37
            self.deltaj0 = 0.97
        else:
            self.cdelta = -1
            self.gamma = -1
            self.deltaj0 = -1


class MexicanHat(DOG):
    """Implements the Mexican hat wavelet class.

    This class inherits the DOG class using m=2.

    """
    def __init__(self):
        self.name = 'Mexican Hat'
        self._set_m(2)




def cwt(signal, dt, dj=1/12, s0=-1, J=-1, wavelet='morlet', freqs=None, pad=True):
    """Continuous wavelet transform of the signal at specified scales.

    Parameters
    ----------
    signal : numpy.ndarray, list
        Input signal array.
    dt : float
        Sampling interval.
    dj : float, optional
        Spacing between discrete scales. Default value is 1/12.
        Smaller values will result in better scale resolution, but
        slower calculation and plot.
    s0 : float, optional
        Smallest scale of the wavelet. Default value is 2*dt.
    J : float, optional
        Number of scales less one. Scales range from s0 up to
        s0 * 2**(J * dj), which gives a total of (J + 1) scales.
        Default is J = (log2(N * dt / so)) / dj.
    wavelet : instance of Wavelet class, or string
        Mother wavelet class. Default is Morlet wavelet.
    freqs : numpy.ndarray, optional
        Custom frequencies to use instead of the ones corresponding
        to the scales described above. Corresponding scales are
        calculated using the wavelet Fourier wavelength.
    pad : optional. Default is True. if set, then pad the time series 
        with enough zeroes to get N up to the next higher power of 2. 
        This prevents wraparound from the end of the time series to 
        the beginning, and also speeds up the FFT's used to do the 
        wavelet transform. This will not eliminate all edge effects.
        (added by SJ)

    Returns
    -------
    W : numpy.ndarray
        Wavelet transform according to the selected mother wavelet.
        Has (J+1) x N dimensions.
    sj : numpy.ndarray
        Vector of scale indices given by sj = s0 * 2**(j * dj),
        j={0, 1, ..., J}.
    freqs : array like
        Vector of Fourier frequencies (in 1 / time units) that
        corresponds to the wavelet scales.
    coi : numpy.ndarray
        Returns the cone of influence, which is a vector of N
        points containing the maximum Fourier period of useful
        information at that particular time. Periods greater than
        those are subject to edge effects.
    fft : numpy.ndarray
        Normalized fast Fourier transform of the input signal.
    fftfreqs : numpy.ndarray
        Fourier frequencies (in 1/time units) for the calculated
        FFT spectrum.

    Example
    -------
    >> mother = wavelet.Morlet(6.)
    >> wave, scales, freqs, coi, fft, fftfreqs = wavelet.cwt(signal,
           0.25, 0.25, 0.5, 28, mother)

    """
    wavelet = _check_parameter_wavelet(wavelet)

    # Original signal length
    n0 = len(signal)
    # If no custom frequencies are set, then set default frequencies
    # according to input parameters `dj`, `s0` and `J`. Otherwise, set wavelet
    # scales according to Fourier equivalent frequencies.
    
    # Zero-padding if enabled (added + the entire code further modified by SJ)
    if pad:
        next_power_of_two = int(2 ** np.ceil(np.log2(n0)))
        signal_padded = np.zeros(next_power_of_two)
        signal_padded[:n0] = signal - np.mean(signal)  # Remove the mean and pad with zeros
    else:
        signal_padded = signal - np.mean(signal)  # Remove the mean without padding
    
    N = len(signal_padded)  # Length of the padded signal

    # Calculate scales and frequencies if not provided
    if freqs is None:
        # Smallest resolvable scale
        # if s0 == -1:
        #     s0 = 2 * dt / wavelet.flambda()
        # # Number of scales
        # if J == -1:
        #     J = int(np.round(np.log2(N * dt / s0) / dj))
        if s0 == -1:
            s0 = 2 * dt
        if J == -1:
            J = int((np.log(float(N) * dt / s0) / np.log(2)) / dj)
        # The scales as of Mallat 1999
        sj = s0 * 2 ** (np.arange(0, J + 1) * dj)
        # Fourier equivalent frequencies
        freqs = 1 / (wavelet.flambda() * sj)
    else:
        # The wavelet scales using custom frequencies.
        sj = 1 / (wavelet.flambda() * freqs)

    # Fourier transform of the (padded) signal
    signal_ft = fft.fft(signal_padded, **fft_kwargs(signal_padded))
    # Fourier angular frequencies
    ftfreqs = 2 * np.pi * fft.fftfreq(N, dt)

    # Creates wavelet transform matrix as outer product of scaled transformed
    # wavelets and transformed signal according to the convolution theorem.
    # (i)   Transform scales to column vector for outer product;
    # (ii)  Calculate 2D matrix [s, f] for each scale s and Fourier angular
    #       frequency f;
    # (iii) Calculate wavelet transform;
    sj_col = sj[:, np.newaxis]
    psi_ft_bar = ((sj_col * ftfreqs[1] * N) ** .5 *
                  np.conjugate(wavelet.psi_ft(sj_col * ftfreqs)))
    W = fft.ifft(signal_ft * psi_ft_bar, axis=1,
                 **fft_kwargs(signal_ft, overwrite_x=True))

    # Trim the wavelet transform to original signal length if padded
    if pad:
        W = W[:, :n0]  # Trim to the original signal length
    
    # Checks for NaN in transform results and removes them from the scales if
    # needed, frequencies and wavelet transform. Trims wavelet transform at
    # length `n0`.
    sel = np.invert(np.isnan(W).all(axis=1))
    if np.any(sel):
        sj = sj[sel]
        freqs = freqs[sel]
        W = W[sel, :]

    # Determines the cone-of-influence. Note that it is returned as a function
    # of time in Fourier periods. Uses triangualr Bartlett window with
    # non-zero end-points.
    coi = (n0 / 2 - np.abs(np.arange(0, n0) - (n0 - 1) / 2))
    coi = wavelet.flambda() * wavelet.coi() * dt * coi

    return (W[:, :n0], sj, freqs, coi, signal_ft[1:N//2] / N ** 0.5,
            ftfreqs[1:N//2] / (2 * np.pi))


def icwt(W, sj, dt, dj=1/12, wavelet='morlet'):
    """Inverse continuous wavelet transform.

    Parameters
    ----------
    W : numpy.ndarray
        Wavelet transform, the result of the `cwt` function.
    sj : numpy.ndarray
        Vector of scale indices as returned by the `cwt` function.
    dt : float
        Sample spacing.
    dj : float, optional
        Spacing between discrete scales as used in the `cwt`
        function. Default value is 0.25.
    wavelet : instance of Wavelet class, or string
        Mother wavelet class. Default is Morlet

    Returns
    -------
    iW : numpy.ndarray
        Inverse wavelet transform.

    Example
    -------
    >> mother = wavelet.Morlet()
    >> wave, scales, freqs, coi, fft, fftfreqs = wavelet.cwt(var,
           0.25, 0.25, 0.5, 28, mother)
    >> iwave = wavelet.icwt(wave, scales, 0.25, 0.25, mother)

    """
    wavelet = _check_parameter_wavelet(wavelet)

    a, b = W.shape
    c = sj.size
    if a == c:
        sj = (np.ones([b, 1]) * sj).transpose()
    elif b == c:
        sj = np.ones([a, 1]) * sj
    else:
        raise Warning('Input array dimensions do not match.')

    # As of Torrence and Compo (1998), eq. (11)
    iW = (dj * np.sqrt(dt) / (wavelet.cdelta * wavelet.psi(0)) *
          (np.real(W) / np.sqrt(sj)).sum(axis=0))
    return iW


def significance(signal, dt, scales, sigma_test=0, alpha=None,
                 significance_level=0.95, dof=-1, wavelet='morlet'):
    """Significance test for the one dimensional wavelet transform.

    Parameters
    ----------
    signal : array like, float
        Input signal array. If a float number is given, then the
        variance is assumed to have this value. If an array is
        given, then its variance is automatically computed.
    dt : float
        Sample spacing.
    scales : array like
        Vector of scale indices given returned by `cwt` function.
    sigma_test : int, optional
        Sets the type of significance test to be performed.
        Accepted values are 0 (default), 1 or 2. See notes below for
        further details.
    alpha : float, optional
        Lag-1 autocorrelation, used for the significance levels.
        Default is 0.0.
    significance_level : float, optional
        Significance level to use. Default is 0.95.
    dof : variant, optional
        Degrees of freedom for significance test to be set
        according to the type set in sigma_test.
    wavelet : instance of Wavelet class, or string
        Mother wavelet class. Default is Morlet

    Returns
    -------
    signif : array like
        Significance levels as a function of scale.
    fft_theor (array like):
        Theoretical red-noise spectrum as a function of period.

    Notes
    -----
    If sigma_test is set to 0, performs a regular chi-square test,
    according to Torrence and Compo (1998) equation 18.

    If set to 1, performs a time-average test (equation 23). In
    this case, dof should be set to the number of local wavelet
    spectra that where averaged together. For the global
    wavelet spectra it would be dof=N, the number of points in
    the time-series.

    If set to 2, performs a scale-average test (equations 25 to
    28). In this case dof should be set to a two element vector
    [s1, s2], which gives the scale range that were averaged
    together. If, for example, the average between scales 2 and
    8 was taken, then dof=[2, 8].

    """
    wavelet = _check_parameter_wavelet(wavelet)

    try:
        n0 = len(signal)
    except TypeError:
        n0 = 1
    J = len(scales) - 1
    dj = np.log2(scales[1] / scales[0])

    if n0 == 1:
        variance = signal
    else:
        variance = signal.std() ** 2

    if alpha is None:
        alpha, _, _ = ar1(signal)

    period = scales * wavelet.flambda()  # Fourier equivalent periods
    freq = dt / period                   # Normalized frequency
    dofmin = wavelet.dofmin              # Degrees of freedom with no smoothing
    Cdelta = wavelet.cdelta              # Reconstruction factor
    gamma_fac = wavelet.gamma            # Time-decorrelation factor
    dj0 = wavelet.deltaj0                # Scale-decorrelation factor

    # Theoretical discrete Fourier power spectrum of the noise signal
    # following Gilman et al. (1963) and Torrence and Compo (1998),
    # equation 16.
    def pk(k, a, N):
        return (1 - a ** 2) / (1 + a ** 2 - 2 * a * np.cos(2 * np.pi * k / N))
    fft_theor = pk(freq, alpha, n0)
    fft_theor = variance * fft_theor     # Including time-series variance
    signif = fft_theor

    try:
        if dof == -1:
            dof = dofmin
    except ValueError:
        pass

    if sigma_test == 0:  # No smoothing, dof=dofmin, TC98 sec. 4
        dof = dofmin
        # As in Torrence and Compo (1998), equation 18.
        chisquare = chi2.ppf(significance_level, dof) / dof
        signif = fft_theor * chisquare
    elif sigma_test == 1:  # Time-averaged significance
        if len(dof) == 1:
            dof = np.zeros(1, J+1) + dof
        sel = find(dof < 1)
        dof[sel] = 1
        # As in Torrence and Compo (1998), equation 23:
        dof = dofmin * (1 + (dof * dt / gamma_fac / scales) ** 2) ** 0.5
        sel = find(dof < dofmin)
        dof[sel] = dofmin  # Minimum dof is dofmin
        for n, d in enumerate(dof):
            chisquare = chi2.ppf(significance_level, d) / d
            signif[n] = fft_theor[n] * chisquare
    elif sigma_test == 2:  # Time-averaged significance
        if len(dof) != 2:
            raise Exception('DOF must be set to [s1, s2], '
                            'the range of scale-averages')
        if Cdelta == -1:
            raise ValueError('Cdelta and dj0 not defined '
                             'for {} with f0={}'.format(wavelet.name,
                                                        wavelet.f0))
        s1, s2 = dof
        sel = find((scales >= s1) & (scales <= s2))
        navg = sel.size
        if navg == 0:
            raise ValueError('No valid scales between {} and {}.'.format(s1,
                                                                         s2))
        # As in Torrence and Compo (1998), equation 25.
        Savg = 1 / sum(1. / scales[sel])
        # Power-of-two mid point:
        Smid = np.exp((np.log(s1) + np.log(s2)) / 2.)
        # As in Torrence and Compo (1998), equation 28.
        dof = (dofmin * navg * Savg / Smid) * \
              ((1 + (navg * dj / dj0) ** 2) ** 0.5)
        # As in Torrence and Compo (1998), equation 27.
        fft_theor = Savg * sum(fft_theor[sel] / scales[sel])
        chisquare = chi2.ppf(significance_level, dof) / dof
        # As in Torrence and Compo (1998), equation 26.
        signif = (dj * dt / Cdelta / Savg) * fft_theor * chisquare
    else:
        raise ValueError('sigma_test must be either 0, 1, or 2.')

    return signif, fft_theor


def xwt(y1, y2, dt, dj=1/12, s0=-1, J=-1, significance_level=0.95,
        wavelet='morlet', normalize=False, no_default_signif=False):
    """Cross wavelet transform (XWT) of two signals.

    The XWT finds regions in time frequency space where the time series
    show high common power.

    Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signal array to calculate cross wavelet transform.
    dt : float
        Sample spacing.
    dj : float, optional
        Spacing between discrete scales. Default value is 1/12.
        Smaller values will result in better scale resolution, but
        slower calculation and plot.
    s0 : float, optional
        Smallest scale of the wavelet. Default value is 2*dt.
    J : float, optional
        Number of scales less one. Scales range from s0 up to
        s0 * 2**(J * dj), which gives a total of (J + 1) scales.
        Default is J = (log2(N*dt/so))/dj.
    wavelet : instance of a wavelet class, optional
        Mother wavelet class. Default is Morlet wavelet.
    significance_level : float, optional
        Significance level to use. Default is 0.95.
    normalize : bool, optional
        If set to true, normalizes CWT by the standard deviation of
        the signals.

    Returns
    -------
    xwt (array like):
        Cross wavelet transform according to the selected mother
        wavelet.
    x (array like):
        Intersected independent variable.
    coi (array like):
        Cone of influence, which is a vector of N points containing
        the maximum Fourier period of useful information at that
        particular time. Periods greater than those are subject to
        edge effects.
    freqs (array like):
        Vector of Fourier equivalent frequencies (in 1 / time units)
        that correspond to the wavelet scales.
    signif (array like):
        Significance levels as a function of scale.

    Notes
    -----
    Torrence and Compo (1998) state that the percent point function
    (PPF) -- inverse of the cumulative distribution function -- of a
    chi-square distribution at 95% confidence and two degrees of
    freedom is Z2(95%)=3.999. However, calculating the PPF using
    chi2.ppf gives Z2(95%)=5.991. To ensure similar significance
    intervals as in Grinsted et al. (2004), one has to use confidence
    of 86.46%.

    """
    wavelet = _check_parameter_wavelet(wavelet)

    # Makes sure input signal are numpy arrays.
    y1 = np.asarray(y1)
    y2 = np.asarray(y2)
    # Calculates the standard deviation of both input signals.
    std1 = y1.std()
    std2 = y2.std()
    # Normalizes both signals, if appropriate.
    if normalize:
        y1_normal = (y1 - y1.mean()) / std1
        y2_normal = (y2 - y2.mean()) / std2
    else:
        y1_normal = y1
        y2_normal = y2

    # Calculates the CWT of the time-series making sure the same parameters
    # are used in both calculations.
    _kwargs = dict(dj=dj, s0=s0, J=J, wavelet=wavelet)
    W1, sj, freq, coi, _, _ = cwt(y1_normal, dt, **_kwargs)
    W2, sj, freq, coi, _, _ = cwt(y2_normal, dt, **_kwargs)

    # Now the wavelet transform coherence
    # W12ini = W1 * W2.conj()
    # scales = np.ones([1, y1.size]) * sj[:, None]
    # # -- Normalization by Scale and Smoothing
    # W12 = wavelet.smooth(W12ini / scales, dt, dj, sj)

    # Calculates the cross CWT of y1 and y2.
    W12 = W1 * W2.conj()

    # And the significance tests. Note that the confidence level is calculated
    # using the percent point function (PPF) of the chi-squared cumulative
    # distribution function (CDF) instead of using Z1(95%) = 2.182 and
    # Z2(95%)=3.999 as suggested by Torrence & Compo (1998) and Grinsted et
    # al. (2004). If the CWT has been normalized, then std1 and std2 should
    # be reset to unity, otherwise the standard deviation of both series have
    # to be calculated.
    if normalize:
        std1 = std2 = 1.
    a1, _, _ = ar1(y1)
    a2, _, _ = ar1(y2)
    Pk1 = ar1_spectrum(freq * dt, a1)
    Pk2 = ar1_spectrum(freq * dt, a2)
    dof = wavelet.dofmin
    if not no_default_signif:
        PPF = chi2.ppf(significance_level, dof)
        signif = (std1 * std2 * (Pk1 * Pk2) ** 0.5 * PPF / dof)
    else:
        signif = np.asarray([0])

    # The resuts:
    return W12, coi, freq, signif


def wct(y1, y2, dt, dj=1/12, s0=-1, J=-1, sig=False,
        significance_level=0.95, wavelet='morlet', normalize=False, **kwargs):
    """Wavelet coherence transform (WCT).

    The WCT finds regions in time frequency space where the two time
    series co-vary, but do not necessarily have high power.

    Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.
    dj : float, optional
        Spacing between discrete scales. Default value is 1/12.
        Smaller values will result in better scale resolution, but
        slower calculation and plot.
    s0 : float, optional
        Smallest scale of the wavelet. Default value is 2*dt.
    J : float, optional
        Number of scales less one. Scales range from s0 up to
        s0 * 2**(J * dj), which gives a total of (J + 1) scales.
        Default is J = (log2(N*dt/so))/dj.
    sig : bool 
        set to compute signficance, default is True
    significance_level (float, optional) :
        Significance level to use. Default is 0.95.
    normalize (boolean, optional) :
        If set to true, normalizes CWT by the standard deviation of
        the signals.

    Returns
    -------
    WCT : magnitude of coherence
    aWCT : phase angle of coherence
    coi (array like):
        Cone of influence, which is a vector of N points containing
        the maximum Fourier period of useful information at that
        particular time. Periods greater than those are subject to
        edge effects.
    freq (array like):
        Vector of Fourier equivalent frequencies (in 1 / time units)    coi :  
    sig :  Significance levels as a function of scale 
       if sig=True when called, otherwise zero.

    See also
    --------
    cwt, xwt

    """
    wavelet = _check_parameter_wavelet(wavelet)

    nt = len(y1)

    # Checking some input parameters
    if s0 == -1:
        # s0 = 2 * dt / wavelet.flambda()
        s0 = 2 * dt
    if J == -1:
        # Number of scales
        # J = int(np.round(np.log2(y1.size * dt / s0) / dj))
        J = int((np.log(float(nt) * dt / s0) / np.log(2)) / dj)

    # Makes sure input signals are numpy arrays.
    y1 = np.asarray(y1)
    y2 = np.asarray(y2)
    # Calculates the standard deviation of both input signals.
    std1 = y1.std()
    std2 = y2.std()
    # Normalizes both signals, if appropriate.
    if normalize:
        y1_normal = (y1 - y1.mean()) / std1
        y2_normal = (y2 - y2.mean()) / std2
    else:
        y1_normal = y1
        y2_normal = y2

    # Calculates the CWT of the time-series making sure the same parameters
    # are used in both calculations.
    _kwargs = dict(dj=dj, s0=s0, J=J, wavelet=wavelet)
    W1, sj, freq, coi, _, _ = cwt(y1_normal, dt, **_kwargs)
    W2, sj, freq, coi, _, _ = cwt(y2_normal, dt, **_kwargs)

    scales1 = np.ones([1, y1.size]) * sj[:, None]
    scales2 = np.ones([1, y2.size]) * sj[:, None]

    # Smooth the wavelet spectra before truncating -- Time Smoothing
    S1 = wavelet.smooth(np.abs(W1) ** 2 / scales1, dt, dj, sj)
    S2 = wavelet.smooth(np.abs(W2) ** 2 / scales2, dt, dj, sj)

    # Now the wavelet transform coherence
    W12 = W1 * W2.conj()
    scales = np.ones([1, y1.size]) * sj[:, None]
    # -- Normalization by Scale and Scale Smoothing
    S12 = wavelet.smooth(W12 / scales, dt, dj, sj)
    WCT = np.abs(S12) ** 2 / (S1 * S2)
    aWCT = np.angle(W12)

    # Calculates the significance using Monte Carlo simulations with 95%
    # confidence as a function of scale.
    if sig:
        a1, b1, c1 = ar1(y1)
        a2, b2, c2 = ar1(y2)

        sig = wct_significance(a1, a2, dt=dt, dj=dj, s0=s0, J=J,
                               significance_level=significance_level,
                               wavelet=wavelet, **kwargs)
    else:
        sig = np.asarray([0])

    return WCT, aWCT, coi, freq, sig


def wct_significance(al1, al2, dt, dj, s0, J, significance_level=0.95,
                     wavelet='morlet', mc_count=50, progress=True,
                     cache=True):
    """Wavelet coherence transform significance.

    Calculates WCT significance using Monte Carlo simulations with
    95% confidence.

    Parameters
    ----------
    al1, al2: float
        Lag-1 autoregressive coeficients of both time series.
    dt : float
        Sample spacing.
    dj : float, optional
        Spacing between discrete scales. Default value is 1/12.
        Smaller values will result in better scale resolution, but
        slower calculation and plot.
    s0 : float, optional
        Smallest scale of the wavelet. Default value is 2*dt.
    J : float, optional
        Number of scales less one. Scales range from s0 up to
        s0 * 2**(J * dj), which gives a total of (J + 1) scales.
        Default is J = (log2(N*dt/so))/dj.
    significance_level : float, optional
        Significance level to use. Default is 0.95.
    wavelet : instance of a wavelet class, optional
        Mother wavelet class. Default is Morlet wavelet.
    mc_count : integer, optional
        Number of Monte Carlo simulations. Default is 300.
    progress : bool, optional
        If `True` (default), shows progress bar on screen.
    cache : bool, optional
        If `True` (default) saves cache to file.

    Returns
    -------
    TODO

    """

    if cache:
        # Load cache if previously calculated. It is assumed that wavelet
        # analysis is performed using the wavelet's default parameters.
        aa = np.round(np.arctanh(np.array([al1, al2]) * 4))
        aa = np.abs(aa) + 0.5 * (aa < 0)
        cache_file = 'wct_sig_{:0.5f}_{:0.5f}_{:0.5f}_{:0.5f}_{:d}_{}'\
            .format(aa[0], aa[1], dj, s0 / dt, J, wavelet.name)
        cache_dir = get_cache_dir()
        try:
            dat = np.loadtxt('{}/{}.gz'.format(cache_dir, cache_file),
                             unpack=True)
            print('NOTE: WCT significance loaded from cache.\n')
            return dat
        except IOError:
            pass

    # Some output to the screen
    print('Calculating wavelet coherence significance')

    # Choose N so that largest scale has at least some part outside the COI
    ms = s0 * (2 ** (J * dj)) / dt
    N = int(np.ceil(ms * 6))
    noise1 = rednoise(N, al1, 1)
    nW1, sj, freq, coi, _, _ = cwt(noise1, dt=dt, dj=dj, s0=s0, J=J,
                                   wavelet=wavelet)

    period = np.ones([1, N]) / freq[:, None]
    coi = np.ones([J + 1, 1]) * coi[None, :]
    outsidecoi = (period <= coi)
    scales = np.ones([1, N]) * sj[:, None]
    sig95 = np.zeros(J + 1)
    maxscale = find(outsidecoi.any(axis=1))[-1]
    sig95[outsidecoi.any(axis=1)] = np.nan

    nbins = 1000
    wlc = np.ma.zeros([J + 1, nbins])
    # Displays progress bar with tqdm
    for _ in tqdm(range(mc_count), disable=not progress):
        # Generates two red-noise signals with lag-1 autoregressive
        # coefficients given by al1 and al2
        noise1 = rednoise(N, al1, 1)
        noise2 = rednoise(N, al2, 1)
        # Calculate the cross wavelet transform of both red-noise signals
        kwargs = dict(dt=dt, dj=dj, s0=s0, J=J, wavelet=wavelet)
        nW1, sj, freq, coi, _, _ = cwt(noise1, **kwargs)
        nW2, sj, freq, coi, _, _ = cwt(noise2, **kwargs)
        nW12 = nW1 * nW2.conj()
        # Smooth wavelet wavelet transforms and calculate wavelet coherence
        # between both signals.
        S1 = wavelet.smooth(np.abs(nW1) ** 2 / scales, dt, dj, sj)
        S2 = wavelet.smooth(np.abs(nW2) ** 2 / scales, dt, dj, sj)
        S12 = wavelet.smooth(nW12 / scales, dt, dj, sj)
        R2 = np.ma.array(np.abs(S12) ** 2 / (S1 * S2), mask=~outsidecoi)
        # Walks through each scale outside the cone of influence and builds a
        # coherence coefficient counter.
        for s in range(maxscale):
            cd = np.floor(R2[s, :] * nbins)
            for j, t in enumerate(cd[~cd.mask]):
                wlc[s, int(t)] += 1

    # After many, many, many Monte Carlo simulations, determine the
    # significance using the coherence coefficient counter percentile.
    wlc.mask = (wlc.data == 0.)
    R2y = (np.arange(nbins) + 0.5) / nbins
    for s in range(maxscale):
        sel = ~wlc[s, :].mask
        P = wlc[s, sel].data.cumsum()
        P = (P - 0.5) / P[-1]
        sig95[s] = np.interp(significance_level, P, R2y[sel])

    if cache:
        # Save the results on cache to avoid to many computations in the future
        np.savetxt('{}/{}.gz'.format(cache_dir, cache_file), sig95)

    # And returns the results
    return sig95


def _check_parameter_wavelet(wavelet):
    mothers = {'morlet': Morlet, 'paul': Paul, 'dog': DOG,
               'mexicanhat': MexicanHat}
    # Checks if input parameter is a string. For backwards
    # compatibility with Python 2 we check if instance is a
    # `str`.
    try:
        if isinstance(wavelet, str):
            return mothers[wavelet]()
    except NameError:
        if isinstance(wavelet, str):
            return mothers[wavelet]()
    # Otherwise, return itself.
    return wavelet
